# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""IDL Parser.

Grammar, parse tree visitor and conversion functions for message definitions in
`IDL`_ format.

.. _IDL: https://www.omg.org/spec/IDL/About-IDL/

"""

from __future__ import annotations

import re
from enum import IntEnum, auto
from pathlib import PurePosixPath as Path
from typing import TYPE_CHECKING, cast

from rosbags.interfaces import Nodetype

from .base import normalize_fieldname, parse_message_definition
from .peg import Visitor, parse_grammar

if TYPE_CHECKING:
    from collections.abc import Generator
    from typing import Literal, TypeAlias

    from rosbags.interfaces.typing import (
        BaseDesc,
        Basename,
        Basetype,
        ConstValue,
        Fielddefs,
        FieldDesc,
        NameDesc,
        Typesdict,
    )

    L: TypeAlias = 'tuple[Literal["LITERAL"], str]'

    LitNode: TypeAlias = 'tuple[Literal[Node.LITERAL], ConstValue]'

    Adec: TypeAlias = 'tuple[Literal[Node.ADECLARATOR], NameDesc, LitNode]'
    Anno: TypeAlias = (
        'tuple[Literal[Node.ANNOTATION], str, list[tuple[NameDesc, LitNode]] | LitNode]'
    )
    Const: TypeAlias = 'tuple[Literal[Node.CONST], tuple[Basetype, str, ConstValue]]'
    Expr: TypeAlias = 'LitNode | NameDesc | ExprUnary | ExprBinary'
    ExprBinary: TypeAlias = 'tuple[Literal[Node.EXPRESSION_BINARY], str, int, int]'
    ExprUnary: TypeAlias = 'tuple[Literal[Node.EXPRESSION_UNARY], str, int]'
    Module: TypeAlias = 'tuple[Literal[Node.MODULE], list[Const], list[Struct]]'
    Struct: TypeAlias = 'tuple[Literal[Node.STRUCT], str, Fielddefs]'

GRAMMAR_IDL = r"""
specification
  = definition+

definition
  = ros2idl_sep
  / macro
  / include
  / module_dcl ';'
  / const_dcl ';'
  / type_dcl ';'

ros2idl_sep
  = r'={80}\nIDL:\s[a-zA-Z][\w]*(/[a-zA-Z][\w]*)*'

macro
  = ifndef
  / define
  / endif

ifndef
  = '#ifndef' r'[a-zA-Z0-9_]+'

define
  = '#define' r'[a-zA-Z0-9_]+'

endif
  = '#endif'

include
  = '#include' include_filename

include_filename
  = '<' r'[^>]+' '>'
  / '"' r'[^"]+' '"'

module_dcl
  = annotation* 'module' identifier '{' definition+ '}'

const_dcl
  = 'const' const_type identifier '=' expression

type_dcl
  = typedef_dcl
  / constr_type_dcl

typedef_dcl
  = 'typedef' type_declarator

type_declarator
  = ( template_type_spec
    / simple_type_spec
    / constr_type_dcl
    ) any_declarators

simple_type_spec
  = base_type_spec
  / scoped_name

template_type_spec
  = sequence_type
  / string_type

sequence_type
  = 'sequence' '<' type_spec ',' expression '>'
  / 'sequence' '<' type_spec '>'

type_spec
  = template_type_spec
  / simple_type_spec

any_declarators
  = any_declarator (',' any_declarator)*

any_declarator
  = array_declarator
  / simple_declarator

constr_type_dcl
  = struct_dcl

struct_dcl
  = struct_def

struct_def
  = annotation* 'struct' identifier '{' member+ '}'

member
  = annotation* type_spec declarators ';'

declarators
  = declarator (',' declarator)*

declarator
  = simple_declarator

simple_declarator
  = identifier

array_declarator
  = identifier fixed_array_size+

fixed_array_size
  = '[' expression ']'

annotation
  = '@' scoped_name ('(' annotation_params ')')?

annotation_params
  = annotation_param (',' annotation_param)*
  / expression

annotation_param
  = identifier '=' expression

const_type
  = base_type_spec
  / string_type
  / scoped_name

base_type_spec
  = float_type
  / integer_type
  / char_type
  / boolean_type

integer_type
  = r'u?int(64|32|16|8)\b'
  / r'(unsigned\s+)?((long\s+)?long|int|short)\b'
  / r'octet\b'

float_type
  = r'(long.double|double|float)\b'

char_type
  = r'char\b'

boolean_type
  = r'boolean\b'

string_type
  = 'string' '<' primary_expr binary_operator expression '>'
  / 'string' '<' primary_expr '>'
  / 'string' '<' unary_operator primary_expr '>'
  / r'string\b'

scoped_name
  = identifier '::' scoped_name
  / ':' scoped_name
  / identifier

identifier
  = r'[a-zA-Z_][a-zA-Z_0-9]*'

expression
  = primary_expr binary_operator expression
  / primary_expr
  / unary_operator primary_expr

primary_expr
  = literal
  / scoped_name
  / '(' expression ')'

binary_operator
  = '|'
  / '^'
  / '&'
  / '<<'
  / '>>'
  / '+'
  / '-'
  / '*'
  / '/'
  / '%'

unary_operator
  = '+'
  / '-'
  / '~'

literal
  = boolean_literal
  / float_literal
  / integer_literal
  / character_literal
  / string_literals

boolean_literal
  = 'TRUE'
  / 'FALSE'

integer_literal
  = hexadecimal_literal
  / octal_literal
  / decimal_literal

decimal_literal
  = r'[-+]?[1-9][0-9]+'
  / r'[-+]?[0-9]'

octal_literal
  = r'[-+]?0[0-7]+'

hexadecimal_literal
  = r'[-+]?0[xX][a-fA-F0-9]+'

float_literal
  = r'[-+]?(([0-9]+\.[0-9]*|\.[0-9]+)([eE][-+]?[0-9]+)?|[0-9]+[eE][-+]?[0-9]+)'

character_literal
  = '\'' r'[a-zA-Z0-9_]' '\''

string_literals
  = string_literal+

string_literal
  = '"' r'(\\"|[^"])*' '"'
"""


def normalize_fieldtype(typename: str, field: FieldDesc) -> FieldDesc:
    """Normalize field typename.

    Args:
        typename: Type name of field owner.
        field: Field definition.

    Returns:
        Normalized fieldtype.

    """
    if field[0] == Nodetype.BASE:
        return field

    ftype, args = field
    ifield = field if ftype == Nodetype.NAME else args[0]

    if ifield[0] == Nodetype.BASE:
        return field

    assert isinstance(ifield, tuple)
    assert ifield[0] == Nodetype.NAME

    name = ifield[1]
    if '/' not in name:
        name = str(Path(typename).parent / name)
    elif name.startswith(f'{Path(typename).parent.name}/'):
        name = str(Path(typename).parent.parent / name)

    ifield = Nodetype.NAME, name

    return ifield if ftype == Nodetype.NAME else (ftype, (ifield, args[1]))  # type: ignore[return-value]


class Node(IntEnum):
    """Parse tree node types."""

    LITERAL = auto()
    MODULE = auto()
    CONST = auto()
    STRUCT = auto()
    ADECLARATOR = auto()
    ANNOTATION = auto()
    EXPRESSION_BINARY = auto()
    EXPRESSION_UNARY = auto()


class VisitorIDL(Visitor):
    """IDL file visitor."""

    RULES = parse_grammar(
        GRAMMAR_IDL,
        re.compile(r'(\s|/[*]([^*]|[*](?!/))*[*]/|//[^\n]*$)+', re.MULTILINE | re.DOTALL),
    )

    def __init__(self) -> None:
        """Initialize."""
        super().__init__()
        self.typedefs: dict[str | Basetype, FieldDesc] = {}

    def visit_specification(
        self,
        children: tuple[
            tuple[
                tuple[Literal[Node.MODULE], list[Const], list[Struct]]
                | tuple[Literal[Node.STRUCT], str, list[tuple[str, FieldDesc]]],
                L,
            ]
            | None
        ],
    ) -> Typesdict:
        """Process start symbol, return only children of modules."""
        structs: dict[str, Fielddefs] = {}
        consts: dict[str, list[tuple[str, Basename, ConstValue]]] = {}
        for item in children:
            if item is None:
                continue
            if item[0][0] == Node.MODULE:
                for csubitem in item[0][1]:
                    assert csubitem[0] == Node.CONST
                    if '_Constants/' in csubitem[1][1]:
                        structname, varname = csubitem[1][1].split('_Constants/')
                        if structname not in consts:
                            consts[structname] = []
                        consts[structname].append(
                            (
                                normalize_fieldname(varname),
                                csubitem[1][0][0],
                                csubitem[1][2],
                            ),
                        )

                for ssubitem in item[0][2]:
                    assert ssubitem[0] == Node.STRUCT
                    structs[ssubitem[1]] = ssubitem[2]
                    if ssubitem[1] not in consts:
                        consts[ssubitem[1]] = []
            elif item[0][0] == Node.STRUCT:
                assert item[1] == ('LITERAL', ';')
                consts[item[0][1]] = []
                structs[item[0][1]] = [
                    (x[0], normalize_fieldtype(item[0][1], x[1])) for x in item[0][2]
                ]
        return {k: (consts[k], v) for k, v in structs.items()}

    def visit_macro(self, _: L | tuple[L, str]) -> None:
        """Process macro, suppress output."""

    def visit_include(self, _: tuple[L, tuple[L, str, L]]) -> None:
        """Process include, suppress output."""

    def visit_module_dcl(
        self,
        children: tuple[
            tuple[Anno, ...],
            L,
            NameDesc,
            L,
            tuple[tuple[Const | Struct | Module | None, L] | None, ...],
            L,
        ],
    ) -> Module:
        """Process module declaration."""
        assert len(children) == 6
        assert children[2][0] == Nodetype.NAME
        name = children[2][1]

        definitions = children[4]
        consts: list[Const] = []
        structs: list[Struct] = []
        for item in definitions:
            if item is None:
                continue
            assert item[1] == ('LITERAL', ';')
            if item[0] is None:
                continue
            subitem = item[0]
            if subitem[0] == Node.CONST:
                consts.append(subitem)
            elif subitem[0] == Node.STRUCT:
                structs.append(subitem)
            else:
                assert subitem[0] == Node.MODULE
                consts += subitem[1]
                structs += subitem[2]

        consts = [(ityp, (typ, f'{name}/{subname}', val)) for ityp, (typ, subname, val) in consts]
        structs = [
            (
                typ,
                f'{name}/{subname}',
                [(x[0], normalize_fieldtype(f'{name}/{subname}', x[1])) for x in rest],
            )
            for typ, subname, rest in structs
        ]

        return Node.MODULE, consts, structs

    def visit_const_dcl(self, children: tuple[L, BaseDesc, NameDesc, L, LitNode]) -> Const:
        """Process const declaration."""
        return Node.CONST, (children[1][1], children[2][1], children[4][1])

    def visit_type_dcl(self, children: Struct | None) -> Struct | None:
        """Process type, pass structs, suppress otherwise."""
        if not children:
            return None
        assert children[0] == Node.STRUCT
        return children

    def visit_typedef_dcl(
        self,
        children: tuple[L, tuple[NameDesc | BaseDesc, tuple[NameDesc | Adec, ...]]],
    ) -> None:
        """Process type declarator, register type mapping in instance typedef dictionary."""
        assert len(children) == 2
        dclchildren = children[1]
        assert len(dclchildren) == 2
        base = self.typedefs.get(dclchildren[0][1], dclchildren[0])
        for declarator in dclchildren[1]:
            if declarator[0] == Node.ADECLARATOR:
                alias = declarator[1][1]
                typ, name = base
                assert isinstance(typ, Nodetype)
                assert isinstance(name, str | tuple)
                count = cast('Adec', declarator)[2][1]
                assert isinstance(count, int)
                value = cast('FieldDesc', (Nodetype.ARRAY, ((typ, name), count)))
            else:
                alias = cast('NameDesc', declarator)[1]
                value = base
            self.typedefs[alias] = value

    def visit_sequence_type(
        self,
        children: tuple[L, L, BaseDesc, L] | tuple[L, L, BaseDesc, L, LitNode, L],
    ) -> tuple[Nodetype, tuple[BaseDesc, int]]:
        """Process sequence type specification."""
        if len(children) == 6:
            typ, count = children[4]
            assert typ == Node.LITERAL
            assert isinstance(count, int)
            return Nodetype.SEQUENCE, (children[2], count)
        return Nodetype.SEQUENCE, (children[2], 0)

    def create_struct_field(
        self,
        parts: tuple[
            FieldDesc,
            tuple[NameDesc, ...],
        ],
    ) -> Generator[tuple[str, FieldDesc], None, None]:
        """Create struct field and expand typedefs."""
        typename, params = parts

        while typename[0] == Nodetype.NAME and typename[1] in self.typedefs:
            typename = self.typedefs[typename[1]]

        yield from ((normalize_fieldname(x[1]), typename) for x in params if x)

    def visit_struct_dcl(
        self,
        children: tuple[
            tuple[Anno, ...],
            L,
            NameDesc,
            L,
            list[tuple[tuple[Anno, ...], FieldDesc, tuple[NameDesc, ...], L]],
            L,
        ],
    ) -> Struct:
        """Process struct declaration."""
        fields = [y for x in children[4] for y in self.create_struct_field(x[1:3])]
        return Node.STRUCT, children[2][1], fields

    def visit_declarators(
        self,
        children: tuple[NameDesc, tuple[tuple[L, NameDesc], ...]],
    ) -> tuple[NameDesc, ...]:
        """Process declarators."""
        return children[0], *(x[1] for x in children[1])

    def visit_any_declarators(
        self,
        children: tuple[
            NameDesc | Adec,
            tuple[tuple[L, NameDesc | Adec], ...],
        ],
    ) -> tuple[NameDesc | Adec, ...]:
        """Process any declarators."""
        return children[0], *(x[1] for x in children[1])

    def visit_array_declarator(
        self,
        children: tuple[NameDesc, tuple[tuple[L, LitNode, L]]],
    ) -> Adec:
        """Process array declarator."""
        return Node.ADECLARATOR, children[0], children[1][0][1]

    def visit_annotation(
        self,
        children: tuple[
            L,
            NameDesc,
            tuple[
                tuple[
                    L,
                    tuple[
                        tuple[NameDesc, L, LitNode],
                        tuple[tuple[L, tuple[NameDesc, L, LitNode]], ...],
                    ]
                    | LitNode
                    | NameDesc,
                    L,
                ]
            ]
            | tuple[()],
        ],
    ) -> Anno:
        """Process annotation."""
        assert len(children) == 3
        assert children[1][0] == Nodetype.NAME
        retparams: list[tuple[NameDesc, LitNode]] | LitNode | NameDesc
        if not children[2]:
            retparams = []
        else:
            params = children[2][0][1]
            if params[0] in (Node.LITERAL, Nodetype.NAME):
                retparams = cast('LitNode', params)
            else:
                head = cast('tuple[NameDesc, L, LitNode]', params[0])
                tail = cast('tuple[tuple[L, tuple[NameDesc, L, LitNode]], ...]', params[1])
                flat = [head, *[x[1:][0] for x in tail]]
                assert all(len(x) == 3 for x in flat)
                retparams = [(x[0], x[2]) for x in flat]
        return Node.ANNOTATION, children[1][1], retparams

    def visit_base_type_spec(self, children: str) -> BaseDesc:
        """Process base type specifier."""
        oname = children
        name = {
            'boolean': 'bool',
            'long double': 'float128',
            'double': 'float64',
            'float': 'float32',
        }.get(oname, oname)
        return Nodetype.BASE, (cast('Basename', name), 0)

    def visit_string_type(self, children: str | tuple[L, L, LitNode, L]) -> BaseDesc:
        """Prrocess string type specifier."""
        if isinstance(children, str):
            return Nodetype.BASE, ('string', 0)

        assert len(children) == 4
        assert isinstance(children[0], tuple)
        assert isinstance(children[2][1], int)
        return Nodetype.BASE, ('string', children[2][1])

    def visit_scoped_name(self, children: NameDesc | tuple[NameDesc, L, NameDesc]) -> NameDesc:
        """Process scoped name."""
        if len(children) == 2:
            return children
        assert len(children) == 3
        return Nodetype.NAME, f'{children[0][1]}/{children[2][1]}'

    def visit_identifier(self, children: str) -> NameDesc:
        """Process identifier."""
        return Nodetype.NAME, children

    def visit_expression(
        self,
        children: LitNode | NameDesc | tuple[L, LitNode] | tuple[LitNode, L, LitNode],
    ) -> Expr:
        """Process expression, literals are assumed to be integers only."""
        if children[0] in {
            Node.LITERAL,
            Nodetype.NAME,
        }:
            assert isinstance(children[1], str | bool | int | float)
            return cast('Expr', children)

        assert isinstance(children[0], tuple)
        if len(children) == 3:
            assert isinstance(children[0][1], int)
            assert isinstance(children[1][1], str)
            assert isinstance(children[2][1], int)
            return Node.EXPRESSION_BINARY, children[1][1], children[0][1], children[2][1]
        assert len(children) == 2
        assert isinstance(children[0][1], str)
        assert isinstance(children[1], tuple)
        assert isinstance(children[1][1], int)
        return Node.EXPRESSION_UNARY, children[0][1], children[1][1]

    def visit_boolean_literal(self, children: str) -> LitNode:
        """Process boolean literal."""
        return Node.LITERAL, children[1] == 'TRUE'

    def visit_float_literal(self, children: str) -> LitNode:
        """Process float literal."""
        return Node.LITERAL, float(children)

    def visit_decimal_literal(self, children: str) -> LitNode:
        """Process decimal integer literal."""
        return Node.LITERAL, int(children)

    def visit_octal_literal(self, children: str) -> LitNode:
        """Process octal integer literal."""
        return Node.LITERAL, int(children, 8)

    def visit_hexadecimal_literal(self, children: str) -> LitNode:
        """Process hexadecimal integer literal."""
        return Node.LITERAL, int(children, 16)

    def visit_character_literal(self, children: tuple[L, str, L]) -> LitNode:
        """Process char literal."""
        return Node.LITERAL, children[1]

    def visit_string_literals(self, children: tuple[tuple[L, str, L], ...]) -> LitNode:
        """Process string literal."""
        return Node.LITERAL, ''.join(x[1] for x in children)


def get_types_from_idl(text: str) -> Typesdict:
    """Get types from idl message definition.

    Args:
        text: Message definition.

    Returns:
        List of message message names and parsetrees.

    """
    return parse_message_definition(VisitorIDL(), text)



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Metadata Tests."""

from __future__ import annotations

from typing import cast

from ruamel.yaml import YAML

from rosbags.interfaces import (
    Qos,
    QosDurability,
    QosHistory,
    QosLiveliness,
    QosReliability,
    QosTime,
)
from rosbags.rosbag2.metadata import dump_qos_v8, dump_qos_v9, parse_qos

REF = [
    Qos(
        QosHistory.UNKNOWN,
        0,
        QosReliability.RELIABLE,
        QosDurability.VOLATILE,
        QosTime(9223372036, 854775807),
        QosTime(9223372036, 854775807),
        QosLiveliness.AUTOMATIC,
        QosTime(9223372036, 854775807),
        avoid_ros_namespace_conventions=False,
    )
]

REF_UNKNOWN = [
    Qos(
        QosHistory.UNKNOWN,
        0,
        QosReliability.UNKNOWN,
        QosDurability.UNKNOWN,
        QosTime(9223372036, 854775807),
        QosTime(9223372036, 854775807),
        QosLiveliness.UNKNOWN,
        QosTime(9223372036, 854775807),
        avoid_ros_namespace_conventions=False,
    )
]

QOS_V8_EMPTY = 'offered_qos_profiles: ""'
QOS_V8 = (
    r'offered_qos_profiles: "- history: 3\n  depth: 0\n  reliability: 1\n  durability: 2\n  '
    r'deadline:\n    sec: 9223372036\n    nsec: 854775807\n  lifespan:\n    sec: 9223372036\n    '
    r'nsec: 854775807\n  liveliness: 1\n  liveliness_lease_duration:\n    sec: 9223372036\n    '
    r'nsec: 854775807\n  avoid_ros_namespace_conventions: false"'
)

QOS_V8_BADENTRIES = (
    r'offered_qos_profiles: "- history: 42\n  depth: 0\n  reliability: 42\n  durability: 42\n  '
    r'deadline:\n    sec: 9223372036\n    nsec: 854775807\n  lifespan:\n    sec: 9223372036\n    '
    r'nsec: 854775807\n  liveliness: 42\n  liveliness_lease_duration:\n    sec: 9223372036\n    '
    r'nsec: 854775807\n  avoid_ros_namespace_conventions: false"'
)


QOS_V9_EMPTY = 'offered_qos_profiles: []'
QOS_V9 = """
offered_qos_profiles:
  - history: unknown
    depth: 0
    reliability: reliable
    durability: volatile
    deadline:
      sec: 9223372036
      nsec: 854775807
    lifespan:
      sec: 9223372036
      nsec: 854775807
    liveliness: automatic
    liveliness_lease_duration:
      sec: 9223372036
      nsec: 854775807
    avoid_ros_namespace_conventions: false
"""


def test_qos_parsing() -> None:
    """Test qos values are parsed correctly."""
    yaml = YAML(typ='safe')
    dct = cast(
        'dict[str, str]',
        yaml.load(QOS_V8_EMPTY),
    )
    assert parse_qos(dct['offered_qos_profiles']) == []

    dct = cast(
        'dict[str, str]',
        yaml.load(QOS_V8),
    )
    assert parse_qos(dct['offered_qos_profiles']) == REF

    dct = cast(
        'dict[str, str]',
        yaml.load(QOS_V9_EMPTY),
    )
    assert parse_qos(dct['offered_qos_profiles']) == []

    dct = cast(
        'dict[str, str]',
        yaml.load(QOS_V9),
    )
    assert parse_qos(dct['offered_qos_profiles']) == REF


def test_qos_v8_out_of_range() -> None:
    """Test qos v8 out of range enums are mapped to unknown."""
    yaml = YAML(typ='safe')
    dct = cast(
        'dict[str, str]',
        yaml.load(QOS_V8_BADENTRIES),
    )
    assert parse_qos(dct['offered_qos_profiles']) == REF_UNKNOWN


def test_qos_roundtrip() -> None:
    """Test qos values stay the same after roundtrip."""
    assert parse_qos(dump_qos_v8([])) == []
    assert parse_qos(dump_qos_v8(REF)) == REF
    assert parse_qos(dump_qos_v9([])) == []
    assert parse_qos(dump_qos_v9(REF)) == REF



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Writer/Reader Roundtrip Tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from rosbags.rosbag2 import CompressionFormat, CompressionMode, Reader, Writer, WriterError
from rosbags.typesys import Stores, get_typestore
from rosbags.typesys.stores.latest import std_msgs__msg__Float64 as Float64

if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize('mode', [*CompressionMode])
def test_roundtrip(mode: CompressionMode, tmp_path: Path) -> None:
    """Test messages stay the same between write and read."""
    store = get_typestore(Stores.LATEST)

    float64 = Float64(1.25)

    path = tmp_path / 'rosbag2'
    wbag = Writer(path, version=Writer.VERSION_LATEST)
    wbag.set_compression(mode, CompressionFormat.ZSTD)
    if mode == CompressionMode.STORAGE:
        with pytest.raises(WriterError, match='storage-side compression'):
            wbag.open()
        return

    with wbag:
        wconnection = wbag.add_connection('/test', float64.__msgtype__, typestore=store)
        wbag.write(wconnection, 42, store.serialize_cdr(float64, float64.__msgtype__))

    rbag = Reader(path)
    with rbag:
        gen = rbag.messages()
        rconnection, _, raw = next(gen)
        assert rconnection.topic == wconnection.topic
        assert rconnection.msgtype == wconnection.msgtype
        assert rconnection.ext == wconnection.ext
        msg = store.deserialize_cdr(raw, rconnection.msgtype)
        assert msg == float64
        with pytest.raises(StopIteration):
            _ = next(gen)



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Rosbag2 errors."""

from __future__ import annotations


class ReaderError(Exception):
    """Reader Error."""

    args: tuple[str]


class WriterError(Exception):
    """Writer Error."""

    args: tuple[str]



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
#
# THIS FILE IS GENERATED, DO NOT EDIT
"""Message type definitions."""

# ruff: noqa: N801,N816,RUF100,TC004

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from rosbags.interfaces import Nodetype as T

from . import ros2_iron as base

if TYPE_CHECKING:
    from typing import ClassVar

    from rosbags.interfaces.typing import Typesdict


action_msgs__msg__GoalInfo = base.action_msgs__msg__GoalInfo
action_msgs__msg__GoalStatus = base.action_msgs__msg__GoalStatus
action_msgs__msg__GoalStatusArray = base.action_msgs__msg__GoalStatusArray
actionlib_msgs__msg__GoalID = base.actionlib_msgs__msg__GoalID
actionlib_msgs__msg__GoalStatus = base.actionlib_msgs__msg__GoalStatus
actionlib_msgs__msg__GoalStatusArray = base.actionlib_msgs__msg__GoalStatusArray
builtin_interfaces__msg__Duration = base.builtin_interfaces__msg__Duration
builtin_interfaces__msg__Time = base.builtin_interfaces__msg__Time
diagnostic_msgs__msg__DiagnosticArray = base.diagnostic_msgs__msg__DiagnosticArray
diagnostic_msgs__msg__DiagnosticStatus = base.diagnostic_msgs__msg__DiagnosticStatus
diagnostic_msgs__msg__KeyValue = base.diagnostic_msgs__msg__KeyValue
geometry_msgs__msg__Accel = base.geometry_msgs__msg__Accel
geometry_msgs__msg__AccelStamped = base.geometry_msgs__msg__AccelStamped
geometry_msgs__msg__AccelWithCovariance = base.geometry_msgs__msg__AccelWithCovariance
geometry_msgs__msg__AccelWithCovarianceStamped = base.geometry_msgs__msg__AccelWithCovarianceStamped
geometry_msgs__msg__Inertia = base.geometry_msgs__msg__Inertia
geometry_msgs__msg__InertiaStamped = base.geometry_msgs__msg__InertiaStamped
geometry_msgs__msg__Point = base.geometry_msgs__msg__Point
geometry_msgs__msg__Point32 = base.geometry_msgs__msg__Point32
geometry_msgs__msg__PointStamped = base.geometry_msgs__msg__PointStamped
geometry_msgs__msg__Polygon = base.geometry_msgs__msg__Polygon
geometry_msgs__msg__PolygonStamped = base.geometry_msgs__msg__PolygonStamped
geometry_msgs__msg__Pose = base.geometry_msgs__msg__Pose
geometry_msgs__msg__Pose2D = base.geometry_msgs__msg__Pose2D
geometry_msgs__msg__PoseArray = base.geometry_msgs__msg__PoseArray
geometry_msgs__msg__PoseStamped = base.geometry_msgs__msg__PoseStamped
geometry_msgs__msg__PoseWithCovariance = base.geometry_msgs__msg__PoseWithCovariance
geometry_msgs__msg__PoseWithCovarianceStamped = base.geometry_msgs__msg__PoseWithCovarianceStamped
geometry_msgs__msg__Quaternion = base.geometry_msgs__msg__Quaternion
geometry_msgs__msg__QuaternionStamped = base.geometry_msgs__msg__QuaternionStamped
geometry_msgs__msg__Transform = base.geometry_msgs__msg__Transform
geometry_msgs__msg__TransformStamped = base.geometry_msgs__msg__TransformStamped
geometry_msgs__msg__Twist = base.geometry_msgs__msg__Twist
geometry_msgs__msg__TwistStamped = base.geometry_msgs__msg__TwistStamped
geometry_msgs__msg__TwistWithCovariance = base.geometry_msgs__msg__TwistWithCovariance
geometry_msgs__msg__TwistWithCovarianceStamped = base.geometry_msgs__msg__TwistWithCovarianceStamped
geometry_msgs__msg__Vector3 = base.geometry_msgs__msg__Vector3
geometry_msgs__msg__Vector3Stamped = base.geometry_msgs__msg__Vector3Stamped
geometry_msgs__msg__Wrench = base.geometry_msgs__msg__Wrench
geometry_msgs__msg__WrenchStamped = base.geometry_msgs__msg__WrenchStamped
lifecycle_msgs__msg__State = base.lifecycle_msgs__msg__State
lifecycle_msgs__msg__Transition = base.lifecycle_msgs__msg__Transition
lifecycle_msgs__msg__TransitionDescription = base.lifecycle_msgs__msg__TransitionDescription
lifecycle_msgs__msg__TransitionEvent = base.lifecycle_msgs__msg__TransitionEvent
nav_msgs__msg__GridCells = base.nav_msgs__msg__GridCells
nav_msgs__msg__MapMetaData = base.nav_msgs__msg__MapMetaData
nav_msgs__msg__OccupancyGrid = base.nav_msgs__msg__OccupancyGrid
nav_msgs__msg__Odometry = base.nav_msgs__msg__Odometry
nav_msgs__msg__Path = base.nav_msgs__msg__Path
rcl_interfaces__msg__FloatingPointRange = base.rcl_interfaces__msg__FloatingPointRange
rcl_interfaces__msg__IntegerRange = base.rcl_interfaces__msg__IntegerRange
rcl_interfaces__msg__ListParametersResult = base.rcl_interfaces__msg__ListParametersResult
rcl_interfaces__msg__LoggerLevel = base.rcl_interfaces__msg__LoggerLevel
rcl_interfaces__msg__Parameter = base.rcl_interfaces__msg__Parameter
rcl_interfaces__msg__ParameterDescriptor = base.rcl_interfaces__msg__ParameterDescriptor
rcl_interfaces__msg__ParameterEvent = base.rcl_interfaces__msg__ParameterEvent
rcl_interfaces__msg__ParameterEventDescriptors = base.rcl_interfaces__msg__ParameterEventDescriptors
rcl_interfaces__msg__ParameterType = base.rcl_interfaces__msg__ParameterType
rcl_interfaces__msg__ParameterValue = base.rcl_interfaces__msg__ParameterValue
rcl_interfaces__msg__SetLoggerLevelsResult = base.rcl_interfaces__msg__SetLoggerLevelsResult
rcl_interfaces__msg__SetParametersResult = base.rcl_interfaces__msg__SetParametersResult
rmw_dds_common__msg__Gid = base.rmw_dds_common__msg__Gid
rmw_dds_common__msg__NodeEntitiesInfo = base.rmw_dds_common__msg__NodeEntitiesInfo
rmw_dds_common__msg__ParticipantEntitiesInfo = base.rmw_dds_common__msg__ParticipantEntitiesInfo
rosgraph_msgs__msg__Clock = base.rosgraph_msgs__msg__Clock
sensor_msgs__msg__BatteryState = base.sensor_msgs__msg__BatteryState
sensor_msgs__msg__CameraInfo = base.sensor_msgs__msg__CameraInfo
sensor_msgs__msg__ChannelFloat32 = base.sensor_msgs__msg__ChannelFloat32
sensor_msgs__msg__CompressedImage = base.sensor_msgs__msg__CompressedImage
sensor_msgs__msg__FluidPressure = base.sensor_msgs__msg__FluidPressure
sensor_msgs__msg__Illuminance = base.sensor_msgs__msg__Illuminance
sensor_msgs__msg__Image = base.sensor_msgs__msg__Image
sensor_msgs__msg__Imu = base.sensor_msgs__msg__Imu
sensor_msgs__msg__JointState = base.sensor_msgs__msg__JointState
sensor_msgs__msg__Joy = base.sensor_msgs__msg__Joy
sensor_msgs__msg__JoyFeedback = base.sensor_msgs__msg__JoyFeedback
sensor_msgs__msg__JoyFeedbackArray = base.sensor_msgs__msg__JoyFeedbackArray
sensor_msgs__msg__LaserEcho = base.sensor_msgs__msg__LaserEcho
sensor_msgs__msg__LaserScan = base.sensor_msgs__msg__LaserScan
sensor_msgs__msg__MagneticField = base.sensor_msgs__msg__MagneticField
sensor_msgs__msg__MultiDOFJointState = base.sensor_msgs__msg__MultiDOFJointState
sensor_msgs__msg__MultiEchoLaserScan = base.sensor_msgs__msg__MultiEchoLaserScan
sensor_msgs__msg__NavSatFix = base.sensor_msgs__msg__NavSatFix
sensor_msgs__msg__PointCloud = base.sensor_msgs__msg__PointCloud
sensor_msgs__msg__PointCloud2 = base.sensor_msgs__msg__PointCloud2
sensor_msgs__msg__PointField = base.sensor_msgs__msg__PointField
sensor_msgs__msg__Range = base.sensor_msgs__msg__Range
sensor_msgs__msg__RegionOfInterest = base.sensor_msgs__msg__RegionOfInterest
sensor_msgs__msg__RelativeHumidity = base.sensor_msgs__msg__RelativeHumidity
sensor_msgs__msg__Temperature = base.sensor_msgs__msg__Temperature
sensor_msgs__msg__TimeReference = base.sensor_msgs__msg__TimeReference
service_msgs__msg__ServiceEventInfo = base.service_msgs__msg__ServiceEventInfo
shape_msgs__msg__Mesh = base.shape_msgs__msg__Mesh
shape_msgs__msg__MeshTriangle = base.shape_msgs__msg__MeshTriangle
shape_msgs__msg__Plane = base.shape_msgs__msg__Plane
shape_msgs__msg__SolidPrimitive = base.shape_msgs__msg__SolidPrimitive
statistics_msgs__msg__MetricsMessage = base.statistics_msgs__msg__MetricsMessage
statistics_msgs__msg__StatisticDataPoint = base.statistics_msgs__msg__StatisticDataPoint
statistics_msgs__msg__StatisticDataType = base.statistics_msgs__msg__StatisticDataType
std_msgs__msg__Bool = base.std_msgs__msg__Bool
std_msgs__msg__Byte = base.std_msgs__msg__Byte
std_msgs__msg__ByteMultiArray = base.std_msgs__msg__ByteMultiArray
std_msgs__msg__Char = base.std_msgs__msg__Char
std_msgs__msg__ColorRGBA = base.std_msgs__msg__ColorRGBA
std_msgs__msg__Empty = base.std_msgs__msg__Empty
std_msgs__msg__Float32 = base.std_msgs__msg__Float32
std_msgs__msg__Float32MultiArray = base.std_msgs__msg__Float32MultiArray
std_msgs__msg__Float64 = base.std_msgs__msg__Float64
std_msgs__msg__Float64MultiArray = base.std_msgs__msg__Float64MultiArray
std_msgs__msg__Header = base.std_msgs__msg__Header
std_msgs__msg__Int16 = base.std_msgs__msg__Int16
std_msgs__msg__Int16MultiArray = base.std_msgs__msg__Int16MultiArray
std_msgs__msg__Int32 = base.std_msgs__msg__Int32
std_msgs__msg__Int32MultiArray = base.std_msgs__msg__Int32MultiArray
std_msgs__msg__Int64 = base.std_msgs__msg__Int64
std_msgs__msg__Int64MultiArray = base.std_msgs__msg__Int64MultiArray
std_msgs__msg__Int8 = base.std_msgs__msg__Int8
std_msgs__msg__Int8MultiArray = base.std_msgs__msg__Int8MultiArray
std_msgs__msg__MultiArrayDimension = base.std_msgs__msg__MultiArrayDimension
std_msgs__msg__MultiArrayLayout = base.std_msgs__msg__MultiArrayLayout
std_msgs__msg__String = base.std_msgs__msg__String
std_msgs__msg__UInt16 = base.std_msgs__msg__UInt16
std_msgs__msg__UInt16MultiArray = base.std_msgs__msg__UInt16MultiArray
std_msgs__msg__UInt32 = base.std_msgs__msg__UInt32
std_msgs__msg__UInt32MultiArray = base.std_msgs__msg__UInt32MultiArray
std_msgs__msg__UInt64 = base.std_msgs__msg__UInt64
std_msgs__msg__UInt64MultiArray = base.std_msgs__msg__UInt64MultiArray
std_msgs__msg__UInt8 = base.std_msgs__msg__UInt8
std_msgs__msg__UInt8MultiArray = base.std_msgs__msg__UInt8MultiArray
stereo_msgs__msg__DisparityImage = base.stereo_msgs__msg__DisparityImage
tf2_msgs__msg__TF2Error = base.tf2_msgs__msg__TF2Error
tf2_msgs__msg__TFMessage = base.tf2_msgs__msg__TFMessage
trajectory_msgs__msg__JointTrajectory = base.trajectory_msgs__msg__JointTrajectory
trajectory_msgs__msg__JointTrajectoryPoint = base.trajectory_msgs__msg__JointTrajectoryPoint
trajectory_msgs__msg__MultiDOFJointTrajectory = base.trajectory_msgs__msg__MultiDOFJointTrajectory
trajectory_msgs__msg__MultiDOFJointTrajectoryPoint = (
    base.trajectory_msgs__msg__MultiDOFJointTrajectoryPoint
)
type_description_interfaces__msg__Field = base.type_description_interfaces__msg__Field
type_description_interfaces__msg__FieldType = base.type_description_interfaces__msg__FieldType
type_description_interfaces__msg__IndividualTypeDescription = (
    base.type_description_interfaces__msg__IndividualTypeDescription
)
type_description_interfaces__msg__KeyValue = base.type_description_interfaces__msg__KeyValue
type_description_interfaces__msg__TypeDescription = (
    base.type_description_interfaces__msg__TypeDescription
)
type_description_interfaces__msg__TypeSource = base.type_description_interfaces__msg__TypeSource
unique_identifier_msgs__msg__UUID = base.unique_identifier_msgs__msg__UUID
visualization_msgs__msg__ImageMarker = base.visualization_msgs__msg__ImageMarker
visualization_msgs__msg__InteractiveMarker = base.visualization_msgs__msg__InteractiveMarker
visualization_msgs__msg__InteractiveMarkerControl = (
    base.visualization_msgs__msg__InteractiveMarkerControl
)
visualization_msgs__msg__InteractiveMarkerFeedback = (
    base.visualization_msgs__msg__InteractiveMarkerFeedback
)
visualization_msgs__msg__InteractiveMarkerInit = base.visualization_msgs__msg__InteractiveMarkerInit
visualization_msgs__msg__InteractiveMarkerPose = base.visualization_msgs__msg__InteractiveMarkerPose
visualization_msgs__msg__InteractiveMarkerUpdate = (
    base.visualization_msgs__msg__InteractiveMarkerUpdate
)
visualization_msgs__msg__MarkerArray = base.visualization_msgs__msg__MarkerArray
visualization_msgs__msg__MenuEntry = base.visualization_msgs__msg__MenuEntry
visualization_msgs__msg__MeshFile = base.visualization_msgs__msg__MeshFile
visualization_msgs__msg__UVCoordinate = base.visualization_msgs__msg__UVCoordinate


@dataclass
class geometry_msgs__msg__PolygonInstance:
    """Class for geometry_msgs/msg/PolygonInstance."""

    polygon: geometry_msgs__msg__Polygon
    id: int
    __msgtype__: ClassVar[str] = 'geometry_msgs/msg/PolygonInstance'


@dataclass
class geometry_msgs__msg__PolygonInstanceStamped:
    """Class for geometry_msgs/msg/PolygonInstanceStamped."""

    header: std_msgs__msg__Header
    polygon: geometry_msgs__msg__PolygonInstance
    __msgtype__: ClassVar[str] = 'geometry_msgs/msg/PolygonInstanceStamped'


@dataclass
class geometry_msgs__msg__VelocityStamped:
    """Class for geometry_msgs/msg/VelocityStamped."""

    header: std_msgs__msg__Header
    body_frame_id: str
    reference_frame_id: str
    velocity: geometry_msgs__msg__Twist
    __msgtype__: ClassVar[str] = 'geometry_msgs/msg/VelocityStamped'


@dataclass
class rcl_interfaces__msg__Log:
    """Class for rcl_interfaces/msg/Log."""

    stamp: builtin_interfaces__msg__Time
    level: int
    name: str
    msg: str
    file: str
    function: str
    line: int
    DEBUG: ClassVar[int] = 10
    INFO: ClassVar[int] = 20
    WARN: ClassVar[int] = 30
    ERROR: ClassVar[int] = 40
    FATAL: ClassVar[int] = 50
    __msgtype__: ClassVar[str] = 'rcl_interfaces/msg/Log'


@dataclass
class rosbag2_interfaces__msg__ReadSplitEvent:
    """Class for rosbag2_interfaces/msg/ReadSplitEvent."""

    closed_file: str
    opened_file: str
    node_name: str
    __msgtype__: ClassVar[str] = 'rosbag2_interfaces/msg/ReadSplitEvent'


@dataclass
class rosbag2_interfaces__msg__WriteSplitEvent:
    """Class for rosbag2_interfaces/msg/WriteSplitEvent."""

    closed_file: str
    opened_file: str
    node_name: str
    __msgtype__: ClassVar[str] = 'rosbag2_interfaces/msg/WriteSplitEvent'


@dataclass
class sensor_msgs__msg__NavSatStatus:
    """Class for sensor_msgs/msg/NavSatStatus."""

    status: int
    service: int
    STATUS_UNKNOWN: ClassVar[int] = -2
    STATUS_NO_FIX: ClassVar[int] = -1
    STATUS_FIX: ClassVar[int] = 0
    STATUS_SBAS_FIX: ClassVar[int] = 1
    STATUS_GBAS_FIX: ClassVar[int] = 2
    SERVICE_UNKNOWN: ClassVar[int] = 0
    SERVICE_GPS: ClassVar[int] = 1
    SERVICE_GLONASS: ClassVar[int] = 2
    SERVICE_COMPASS: ClassVar[int] = 4
    SERVICE_GALILEO: ClassVar[int] = 8
    __msgtype__: ClassVar[str] = 'sensor_msgs/msg/NavSatStatus'


@dataclass
class visualization_msgs__msg__Marker:
    """Class for visualization_msgs/msg/Marker."""

    header: std_msgs__msg__Header
    ns: str
    id: int
    type: int
    action: int
    pose: geometry_msgs__msg__Pose
    scale: geometry_msgs__msg__Vector3
    color: std_msgs__msg__ColorRGBA
    lifetime: builtin_interfaces__msg__Duration
    frame_locked: bool
    points: list[geometry_msgs__msg__Point]
    colors: list[std_msgs__msg__ColorRGBA]
    texture_resource: str
    texture: sensor_msgs__msg__CompressedImage
    uv_coordinates: list[visualization_msgs__msg__UVCoordinate]
    text: str
    mesh_resource: str
    mesh_file: visualization_msgs__msg__MeshFile
    mesh_use_embedded_materials: bool
    ARROW: ClassVar[int] = 0
    CUBE: ClassVar[int] = 1
    SPHERE: ClassVar[int] = 2
    CYLINDER: ClassVar[int] = 3
    LINE_STRIP: ClassVar[int] = 4
    LINE_LIST: ClassVar[int] = 5
    CUBE_LIST: ClassVar[int] = 6
    SPHERE_LIST: ClassVar[int] = 7
    POINTS: ClassVar[int] = 8
    TEXT_VIEW_FACING: ClassVar[int] = 9
    MESH_RESOURCE: ClassVar[int] = 10
    TRIANGLE_LIST: ClassVar[int] = 11
    ARROW_STRIP: ClassVar[int] = 12
    ADD: ClassVar[int] = 0
    MODIFY: ClassVar[int] = 0
    DELETE: ClassVar[int] = 2
    DELETEALL: ClassVar[int] = 3
    __msgtype__: ClassVar[str] = 'visualization_msgs/msg/Marker'


FIELDDEFS: Typesdict = {
    'action_msgs/msg/GoalInfo': base.FIELDDEFS['action_msgs/msg/GoalInfo'],
    'action_msgs/msg/GoalStatus': base.FIELDDEFS['action_msgs/msg/GoalStatus'],
    'action_msgs/msg/GoalStatusArray': base.FIELDDEFS['action_msgs/msg/GoalStatusArray'],
    'actionlib_msgs/msg/GoalID': base.FIELDDEFS['actionlib_msgs/msg/GoalID'],
    'actionlib_msgs/msg/GoalStatus': base.FIELDDEFS['actionlib_msgs/msg/GoalStatus'],
    'actionlib_msgs/msg/GoalStatusArray': base.FIELDDEFS['actionlib_msgs/msg/GoalStatusArray'],
    'builtin_interfaces/msg/Duration': base.FIELDDEFS['builtin_interfaces/msg/Duration'],
    'builtin_interfaces/msg/Time': base.FIELDDEFS['builtin_interfaces/msg/Time'],
    'diagnostic_msgs/msg/DiagnosticArray': base.FIELDDEFS['diagnostic_msgs/msg/DiagnosticArray'],
    'diagnostic_msgs/msg/DiagnosticStatus': base.FIELDDEFS['diagnostic_msgs/msg/DiagnosticStatus'],
    'diagnostic_msgs/msg/KeyValue': base.FIELDDEFS['diagnostic_msgs/msg/KeyValue'],
    'geometry_msgs/msg/Accel': base.FIELDDEFS['geometry_msgs/msg/Accel'],
    'geometry_msgs/msg/AccelStamped': base.FIELDDEFS['geometry_msgs/msg/AccelStamped'],
    'geometry_msgs/msg/AccelWithCovariance': base.FIELDDEFS[
        'geometry_msgs/msg/AccelWithCovariance'
    ],
    'geometry_msgs/msg/AccelWithCovarianceStamped': base.FIELDDEFS[
        'geometry_msgs/msg/AccelWithCovarianceStamped'
    ],
    'geometry_msgs/msg/Inertia': base.FIELDDEFS['geometry_msgs/msg/Inertia'],
    'geometry_msgs/msg/InertiaStamped': base.FIELDDEFS['geometry_msgs/msg/InertiaStamped'],
    'geometry_msgs/msg/Point': base.FIELDDEFS['geometry_msgs/msg/Point'],
    'geometry_msgs/msg/Point32': base.FIELDDEFS['geometry_msgs/msg/Point32'],
    'geometry_msgs/msg/PointStamped': base.FIELDDEFS['geometry_msgs/msg/PointStamped'],
    'geometry_msgs/msg/Polygon': base.FIELDDEFS['geometry_msgs/msg/Polygon'],
    'geometry_msgs/msg/PolygonStamped': base.FIELDDEFS['geometry_msgs/msg/PolygonStamped'],
    'geometry_msgs/msg/Pose': base.FIELDDEFS['geometry_msgs/msg/Pose'],
    'geometry_msgs/msg/Pose2D': base.FIELDDEFS['geometry_msgs/msg/Pose2D'],
    'geometry_msgs/msg/PoseArray': base.FIELDDEFS['geometry_msgs/msg/PoseArray'],
    'geometry_msgs/msg/PoseStamped': base.FIELDDEFS['geometry_msgs/msg/PoseStamped'],
    'geometry_msgs/msg/PoseWithCovariance': base.FIELDDEFS['geometry_msgs/msg/PoseWithCovariance'],
    'geometry_msgs/msg/PoseWithCovarianceStamped': base.FIELDDEFS[
        'geometry_msgs/msg/PoseWithCovarianceStamped'
    ],
    'geometry_msgs/msg/Quaternion': base.FIELDDEFS['geometry_msgs/msg/Quaternion'],
    'geometry_msgs/msg/QuaternionStamped': base.FIELDDEFS['geometry_msgs/msg/QuaternionStamped'],
    'geometry_msgs/msg/Transform': base.FIELDDEFS['geometry_msgs/msg/Transform'],
    'geometry_msgs/msg/TransformStamped': base.FIELDDEFS['geometry_msgs/msg/TransformStamped'],
    'geometry_msgs/msg/Twist': base.FIELDDEFS['geometry_msgs/msg/Twist'],
    'geometry_msgs/msg/TwistStamped': base.FIELDDEFS['geometry_msgs/msg/TwistStamped'],
    'geometry_msgs/msg/TwistWithCovariance': base.FIELDDEFS[
        'geometry_msgs/msg/TwistWithCovariance'
    ],
    'geometry_msgs/msg/TwistWithCovarianceStamped': base.FIELDDEFS[
        'geometry_msgs/msg/TwistWithCovarianceStamped'
    ],
    'geometry_msgs/msg/Vector3': base.FIELDDEFS['geometry_msgs/msg/Vector3'],
    'geometry_msgs/msg/Vector3Stamped': base.FIELDDEFS['geometry_msgs/msg/Vector3Stamped'],
    'geometry_msgs/msg/Wrench': base.FIELDDEFS['geometry_msgs/msg/Wrench'],
    'geometry_msgs/msg/WrenchStamped': base.FIELDDEFS['geometry_msgs/msg/WrenchStamped'],
    'lifecycle_msgs/msg/State': base.FIELDDEFS['lifecycle_msgs/msg/State'],
    'lifecycle_msgs/msg/Transition': base.FIELDDEFS['lifecycle_msgs/msg/Transition'],
    'lifecycle_msgs/msg/TransitionDescription': base.FIELDDEFS[
        'lifecycle_msgs/msg/TransitionDescription'
    ],
    'lifecycle_msgs/msg/TransitionEvent': base.FIELDDEFS['lifecycle_msgs/msg/TransitionEvent'],
    'nav_msgs/msg/GridCells': base.FIELDDEFS['nav_msgs/msg/GridCells'],
    'nav_msgs/msg/MapMetaData': base.FIELDDEFS['nav_msgs/msg/MapMetaData'],
    'nav_msgs/msg/OccupancyGrid': base.FIELDDEFS['nav_msgs/msg/OccupancyGrid'],
    'nav_msgs/msg/Odometry': base.FIELDDEFS['nav_msgs/msg/Odometry'],
    'nav_msgs/msg/Path': base.FIELDDEFS['nav_msgs/msg/Path'],
    'rcl_interfaces/msg/FloatingPointRange': base.FIELDDEFS[
        'rcl_interfaces/msg/FloatingPointRange'
    ],
    'rcl_interfaces/msg/IntegerRange': base.FIELDDEFS['rcl_interfaces/msg/IntegerRange'],
    'rcl_interfaces/msg/ListParametersResult': base.FIELDDEFS[
        'rcl_interfaces/msg/ListParametersResult'
    ],
    'rcl_interfaces/msg/LoggerLevel': base.FIELDDEFS['rcl_interfaces/msg/LoggerLevel'],
    'rcl_interfaces/msg/Parameter': base.FIELDDEFS['rcl_interfaces/msg/Parameter'],
    'rcl_interfaces/msg/ParameterDescriptor': base.FIELDDEFS[
        'rcl_interfaces/msg/ParameterDescriptor'
    ],
    'rcl_interfaces/msg/ParameterEvent': base.FIELDDEFS['rcl_interfaces/msg/ParameterEvent'],
    'rcl_interfaces/msg/ParameterEventDescriptors': base.FIELDDEFS[
        'rcl_interfaces/msg/ParameterEventDescriptors'
    ],
    'rcl_interfaces/msg/ParameterType': base.FIELDDEFS['rcl_interfaces/msg/ParameterType'],
    'rcl_interfaces/msg/ParameterValue': base.FIELDDEFS['rcl_interfaces/msg/ParameterValue'],
    'rcl_interfaces/msg/SetLoggerLevelsResult': base.FIELDDEFS[
        'rcl_interfaces/msg/SetLoggerLevelsResult'
    ],
    'rcl_interfaces/msg/SetParametersResult': base.FIELDDEFS[
        'rcl_interfaces/msg/SetParametersResult'
    ],
    'rmw_dds_common/msg/Gid': base.FIELDDEFS['rmw_dds_common/msg/Gid'],
    'rmw_dds_common/msg/NodeEntitiesInfo': base.FIELDDEFS['rmw_dds_common/msg/NodeEntitiesInfo'],
    'rmw_dds_common/msg/ParticipantEntitiesInfo': base.FIELDDEFS[
        'rmw_dds_common/msg/ParticipantEntitiesInfo'
    ],
    'rosgraph_msgs/msg/Clock': base.FIELDDEFS['rosgraph_msgs/msg/Clock'],
    'sensor_msgs/msg/BatteryState': base.FIELDDEFS['sensor_msgs/msg/BatteryState'],
    'sensor_msgs/msg/CameraInfo': base.FIELDDEFS['sensor_msgs/msg/CameraInfo'],
    'sensor_msgs/msg/ChannelFloat32': base.FIELDDEFS['sensor_msgs/msg/ChannelFloat32'],
    'sensor_msgs/msg/CompressedImage': base.FIELDDEFS['sensor_msgs/msg/CompressedImage'],
    'sensor_msgs/msg/FluidPressure': base.FIELDDEFS['sensor_msgs/msg/FluidPressure'],
    'sensor_msgs/msg/Illuminance': base.FIELDDEFS['sensor_msgs/msg/Illuminance'],
    'sensor_msgs/msg/Image': base.FIELDDEFS['sensor_msgs/msg/Image'],
    'sensor_msgs/msg/Imu': base.FIELDDEFS['sensor_msgs/msg/Imu'],
    'sensor_msgs/msg/JointState': base.FIELDDEFS['sensor_msgs/msg/JointState'],
    'sensor_msgs/msg/Joy': base.FIELDDEFS['sensor_msgs/msg/Joy'],
    'sensor_msgs/msg/JoyFeedback': base.FIELDDEFS['sensor_msgs/msg/JoyFeedback'],
    'sensor_msgs/msg/JoyFeedbackArray': base.FIELDDEFS['sensor_msgs/msg/JoyFeedbackArray'],
    'sensor_msgs/msg/LaserEcho': base.FIELDDEFS['sensor_msgs/msg/LaserEcho'],
    'sensor_msgs/msg/LaserScan': base.FIELDDEFS['sensor_msgs/msg/LaserScan'],
    'sensor_msgs/msg/MagneticField': base.FIELDDEFS['sensor_msgs/msg/MagneticField'],
    'sensor_msgs/msg/MultiDOFJointState': base.FIELDDEFS['sensor_msgs/msg/MultiDOFJointState'],
    'sensor_msgs/msg/MultiEchoLaserScan': base.FIELDDEFS['sensor_msgs/msg/MultiEchoLaserScan'],
    'sensor_msgs/msg/NavSatFix': base.FIELDDEFS['sensor_msgs/msg/NavSatFix'],
    'sensor_msgs/msg/PointCloud': base.FIELDDEFS['sensor_msgs/msg/PointCloud'],
    'sensor_msgs/msg/PointCloud2': base.FIELDDEFS['sensor_msgs/msg/PointCloud2'],
    'sensor_msgs/msg/PointField': base.FIELDDEFS['sensor_msgs/msg/PointField'],
    'sensor_msgs/msg/Range': base.FIELDDEFS['sensor_msgs/msg/Range'],
    'sensor_msgs/msg/RegionOfInterest': base.FIELDDEFS['sensor_msgs/msg/RegionOfInterest'],
    'sensor_msgs/msg/RelativeHumidity': base.FIELDDEFS['sensor_msgs/msg/RelativeHumidity'],
    'sensor_msgs/msg/Temperature': base.FIELDDEFS['sensor_msgs/msg/Temperature'],
    'sensor_msgs/msg/TimeReference': base.FIELDDEFS['sensor_msgs/msg/TimeReference'],
    'service_msgs/msg/ServiceEventInfo': base.FIELDDEFS['service_msgs/msg/ServiceEventInfo'],
    'shape_msgs/msg/Mesh': base.FIELDDEFS['shape_msgs/msg/Mesh'],
    'shape_msgs/msg/MeshTriangle': base.FIELDDEFS['shape_msgs/msg/MeshTriangle'],
    'shape_msgs/msg/Plane': base.FIELDDEFS['shape_msgs/msg/Plane'],
    'shape_msgs/msg/SolidPrimitive': base.FIELDDEFS['shape_msgs/msg/SolidPrimitive'],
    'statistics_msgs/msg/MetricsMessage': base.FIELDDEFS['statistics_msgs/msg/MetricsMessage'],
    'statistics_msgs/msg/StatisticDataPoint': base.FIELDDEFS[
        'statistics_msgs/msg/StatisticDataPoint'
    ],
    'statistics_msgs/msg/StatisticDataType': base.FIELDDEFS[
        'statistics_msgs/msg/StatisticDataType'
    ],
    'std_msgs/msg/Bool': base.FIELDDEFS['std_msgs/msg/Bool'],
    'std_msgs/msg/Byte': base.FIELDDEFS['std_msgs/msg/Byte'],
    'std_msgs/msg/ByteMultiArray': base.FIELDDEFS['std_msgs/msg/ByteMultiArray'],
    'std_msgs/msg/Char': base.FIELDDEFS['std_msgs/msg/Char'],
    'std_msgs/msg/ColorRGBA': base.FIELDDEFS['std_msgs/msg/ColorRGBA'],
    'std_msgs/msg/Empty': base.FIELDDEFS['std_msgs/msg/Empty'],
    'std_msgs/msg/Float32': base.FIELDDEFS['std_msgs/msg/Float32'],
    'std_msgs/msg/Float32MultiArray': base.FIELDDEFS['std_msgs/msg/Float32MultiArray'],
    'std_msgs/msg/Float64': base.FIELDDEFS['std_msgs/msg/Float64'],
    'std_msgs/msg/Float64MultiArray': base.FIELDDEFS['std_msgs/msg/Float64MultiArray'],
    'std_msgs/msg/Header': base.FIELDDEFS['std_msgs/msg/Header'],
    'std_msgs/msg/Int16': base.FIELDDEFS['std_msgs/msg/Int16'],
    'std_msgs/msg/Int16MultiArray': base.FIELDDEFS['std_msgs/msg/Int16MultiArray'],
    'std_msgs/msg/Int32': base.FIELDDEFS['std_msgs/msg/Int32'],
    'std_msgs/msg/Int32MultiArray': base.FIELDDEFS['std_msgs/msg/Int32MultiArray'],
    'std_msgs/msg/Int64': base.FIELDDEFS['std_msgs/msg/Int64'],
    'std_msgs/msg/Int64MultiArray': base.FIELDDEFS['std_msgs/msg/Int64MultiArray'],
    'std_msgs/msg/Int8': base.FIELDDEFS['std_msgs/msg/Int8'],
    'std_msgs/msg/Int8MultiArray': base.FIELDDEFS['std_msgs/msg/Int8MultiArray'],
    'std_msgs/msg/MultiArrayDimension': base.FIELDDEFS['std_msgs/msg/MultiArrayDimension'],
    'std_msgs/msg/MultiArrayLayout': base.FIELDDEFS['std_msgs/msg/MultiArrayLayout'],
    'std_msgs/msg/String': base.FIELDDEFS['std_msgs/msg/String'],
    'std_msgs/msg/UInt16': base.FIELDDEFS['std_msgs/msg/UInt16'],
    'std_msgs/msg/UInt16MultiArray': base.FIELDDEFS['std_msgs/msg/UInt16MultiArray'],
    'std_msgs/msg/UInt32': base.FIELDDEFS['std_msgs/msg/UInt32'],
    'std_msgs/msg/UInt32MultiArray': base.FIELDDEFS['std_msgs/msg/UInt32MultiArray'],
    'std_msgs/msg/UInt64': base.FIELDDEFS['std_msgs/msg/UInt64'],
    'std_msgs/msg/UInt64MultiArray': base.FIELDDEFS['std_msgs/msg/UInt64MultiArray'],
    'std_msgs/msg/UInt8': base.FIELDDEFS['std_msgs/msg/UInt8'],
    'std_msgs/msg/UInt8MultiArray': base.FIELDDEFS['std_msgs/msg/UInt8MultiArray'],
    'stereo_msgs/msg/DisparityImage': base.FIELDDEFS['stereo_msgs/msg/DisparityImage'],
    'tf2_msgs/msg/TF2Error': base.FIELDDEFS['tf2_msgs/msg/TF2Error'],
    'tf2_msgs/msg/TFMessage': base.FIELDDEFS['tf2_msgs/msg/TFMessage'],
    'trajectory_msgs/msg/JointTrajectory': base.FIELDDEFS['trajectory_msgs/msg/JointTrajectory'],
    'trajectory_msgs/msg/JointTrajectoryPoint': base.FIELDDEFS[
        'trajectory_msgs/msg/JointTrajectoryPoint'
    ],
    'trajectory_msgs/msg/MultiDOFJointTrajectory': base.FIELDDEFS[
        'trajectory_msgs/msg/MultiDOFJointTrajectory'
    ],
    'trajectory_msgs/msg/MultiDOFJointTrajectoryPoint': base.FIELDDEFS[
        'trajectory_msgs/msg/MultiDOFJointTrajectoryPoint'
    ],
    'type_description_interfaces/msg/Field': base.FIELDDEFS[
        'type_description_interfaces/msg/Field'
    ],
    'type_description_interfaces/msg/FieldType': base.FIELDDEFS[
        'type_description_interfaces/msg/FieldType'
    ],
    'type_description_interfaces/msg/IndividualTypeDescription': base.FIELDDEFS[
        'type_description_interfaces/msg/IndividualTypeDescription'
    ],
    'type_description_interfaces/msg/KeyValue': base.FIELDDEFS[
        'type_description_interfaces/msg/KeyValue'
    ],
    'type_description_interfaces/msg/TypeDescription': base.FIELDDEFS[
        'type_description_interfaces/msg/TypeDescription'
    ],
    'type_description_interfaces/msg/TypeSource': base.FIELDDEFS[
        'type_description_interfaces/msg/TypeSource'
    ],
    'unique_identifier_msgs/msg/UUID': base.FIELDDEFS['unique_identifier_msgs/msg/UUID'],
    'visualization_msgs/msg/ImageMarker': base.FIELDDEFS['visualization_msgs/msg/ImageMarker'],
    'visualization_msgs/msg/InteractiveMarker': base.FIELDDEFS[
        'visualization_msgs/msg/InteractiveMarker'
    ],
    'visualization_msgs/msg/InteractiveMarkerControl': base.FIELDDEFS[
        'visualization_msgs/msg/InteractiveMarkerControl'
    ],
    'visualization_msgs/msg/InteractiveMarkerFeedback': base.FIELDDEFS[
        'visualization_msgs/msg/InteractiveMarkerFeedback'
    ],
    'visualization_msgs/msg/InteractiveMarkerInit': base.FIELDDEFS[
        'visualization_msgs/msg/InteractiveMarkerInit'
    ],
    'visualization_msgs/msg/InteractiveMarkerPose': base.FIELDDEFS[
        'visualization_msgs/msg/InteractiveMarkerPose'
    ],
    'visualization_msgs/msg/InteractiveMarkerUpdate': base.FIELDDEFS[
        'visualization_msgs/msg/InteractiveMarkerUpdate'
    ],
    'visualization_msgs/msg/MarkerArray': base.FIELDDEFS['visualization_msgs/msg/MarkerArray'],
    'visualization_msgs/msg/MenuEntry': base.FIELDDEFS['visualization_msgs/msg/MenuEntry'],
    'visualization_msgs/msg/MeshFile': base.FIELDDEFS['visualization_msgs/msg/MeshFile'],
    'visualization_msgs/msg/UVCoordinate': base.FIELDDEFS['visualization_msgs/msg/UVCoordinate'],
    'geometry_msgs/msg/PolygonInstance': (
        [],
        [
            ('polygon', (T.NAME, 'geometry_msgs/msg/Polygon')),
            ('id', (T.BASE, ('int64', 0))),
        ],
    ),
    'geometry_msgs/msg/PolygonInstanceStamped': (
        [],
        [
            ('header', (T.NAME, 'std_msgs/msg/Header')),
            ('polygon', (T.NAME, 'geometry_msgs/msg/PolygonInstance')),
        ],
    ),
    'geometry_msgs/msg/VelocityStamped': (
        [],
        [
            ('header', (T.NAME, 'std_msgs/msg/Header')),
            ('body_frame_id', (T.BASE, ('string', 0))),
            ('reference_frame_id', (T.BASE, ('string', 0))),
            ('velocity', (T.NAME, 'geometry_msgs/msg/Twist')),
        ],
    ),
    'rcl_interfaces/msg/Log': (
        [
            ('DEBUG', 'uint8', 10),
            ('INFO', 'uint8', 20),
            ('WARN', 'uint8', 30),
            ('ERROR', 'uint8', 40),
            ('FATAL', 'uint8', 50),
        ],
        [
            ('stamp', (T.NAME, 'builtin_interfaces/msg/Time')),
            ('level', (T.BASE, ('uint8', 0))),
            ('name', (T.BASE, ('string', 0))),
            ('msg', (T.BASE, ('string', 0))),
            ('file', (T.BASE, ('string', 0))),
            ('function', (T.BASE, ('string', 0))),
            ('line', (T.BASE, ('uint32', 0))),
        ],
    ),
    'rosbag2_interfaces/msg/ReadSplitEvent': (
        [],
        [
            ('closed_file', (T.BASE, ('string', 0))),
            ('opened_file', (T.BASE, ('string', 0))),
            ('node_name', (T.BASE, ('string', 0))),
        ],
    ),
    'rosbag2_interfaces/msg/WriteSplitEvent': (
        [],
        [
            ('closed_file', (T.BASE, ('string', 0))),
            ('opened_file', (T.BASE, ('string', 0))),
            ('node_name', (T.BASE, ('string', 0))),
        ],
    ),
    'sensor_msgs/msg/NavSatStatus': (
        [
            ('STATUS_UNKNOWN', 'int8', -2),
            ('STATUS_NO_FIX', 'int8', -1),
            ('STATUS_FIX', 'int8', 0),
            ('STATUS_SBAS_FIX', 'int8', 1),
            ('STATUS_GBAS_FIX', 'int8', 2),
            ('SERVICE_UNKNOWN', 'uint16', 0),
            ('SERVICE_GPS', 'uint16', 1),
            ('SERVICE_GLONASS', 'uint16', 2),
            ('SERVICE_COMPASS', 'uint16', 4),
            ('SERVICE_GALILEO', 'uint16', 8),
        ],
        [
            ('status', (T.BASE, ('int8', 0))),
            ('service', (T.BASE, ('uint16', 0))),
        ],
    ),
    'visualization_msgs/msg/Marker': (
        [
            ('ARROW', 'int32', 0),
            ('CUBE', 'int32', 1),
            ('SPHERE', 'int32', 2),
            ('CYLINDER', 'int32', 3),
            ('LINE_STRIP', 'int32', 4),
            ('LINE_LIST', 'int32', 5),
            ('CUBE_LIST', 'int32', 6),
            ('SPHERE_LIST', 'int32', 7),
            ('POINTS', 'int32', 8),
            ('TEXT_VIEW_FACING', 'int32', 9),
            ('MESH_RESOURCE', 'int32', 10),
            ('TRIANGLE_LIST', 'int32', 11),
            ('ARROW_STRIP', 'int32', 12),
            ('ADD', 'int32', 0),
            ('MODIFY', 'int32', 0),
            ('DELETE', 'int32', 2),
            ('DELETEALL', 'int32', 3),
        ],
        [
            ('header', (T.NAME, 'std_msgs/msg/Header')),
            ('ns', (T.BASE, ('string', 0))),
            ('id', (T.BASE, ('int32', 0))),
            ('type', (T.BASE, ('int32', 0))),
            ('action', (T.BASE, ('int32', 0))),
            ('pose', (T.NAME, 'geometry_msgs/msg/Pose')),
            ('scale', (T.NAME, 'geometry_msgs/msg/Vector3')),
            ('color', (T.NAME, 'std_msgs/msg/ColorRGBA')),
            ('lifetime', (T.NAME, 'builtin_interfaces/msg/Duration')),
            ('frame_locked', (T.BASE, ('bool', 0))),
            ('points', (T.SEQUENCE, ((T.NAME, 'geometry_msgs/msg/Point'), 0))),
            ('colors', (T.SEQUENCE, ((T.NAME, 'std_msgs/msg/ColorRGBA'), 0))),
            ('texture_resource', (T.BASE, ('string', 0))),
            ('texture', (T.NAME, 'sensor_msgs/msg/CompressedImage')),
            ('uv_coordinates', (T.SEQUENCE, ((T.NAME, 'visualization_msgs/msg/UVCoordinate'), 0))),
            ('text', (T.BASE, ('string', 0))),
            ('mesh_resource', (T.BASE, ('string', 0))),
            ('mesh_file', (T.NAME, 'visualization_msgs/msg/MeshFile')),
            ('mesh_use_embedded_materials', (T.BASE, ('bool', 0))),
        ],
    ),
}



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Rosbag2 writer."""

from __future__ import annotations

from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING

import zstandard
from ruamel.yaml import YAML

from rosbags.interfaces import (
    Connection,
    ConnectionExtRosbag2,
    MessageDefinition,
    MessageDefinitionFormat,
    Qos,
)

from .enums import CompressionFormat, CompressionMode, StoragePlugin
from .errors import WriterError
from .metadata import dump_qos_v8, dump_qos_v9
from .storage_mcap import McapWriter
from .storage_sqlite3 import Sqlite3Writer

if TYPE_CHECKING:
    import sys
    from collections.abc import Mapping, Sequence
    from types import TracebackType
    from typing import Literal, Protocol

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

    from rosbags.typesys.store import Typestore

    from .metadata import Metadata

    class StorageWriter(Protocol):
        """Storage Writer Protocol."""

        path: Path

        def __init__(self, path: Path, compression: CompressionMode) -> None:
            """Initialize."""
            raise NotImplementedError

        def add_msgtype(self, connection: Connection) -> None:
            """Add a msgtypen."""
            raise NotImplementedError

        def add_connection(self, connection: Connection, offered_qos_profiles: str) -> None:
            """Add a connection."""
            raise NotImplementedError

        def write(self, connection: Connection, timestamp: int, data: bytes | memoryview) -> None:
            """Write message to rosbag2."""

        def close(self, version: int, metadata: str) -> None:
            """Close rosbag2 after writing."""


class Writer:
    """Rosbag2 writer.

    This class implements writing of rosbag2 files in version 8 or 9. It
    should be used as a contextmanager.

    """

    VERSION_LATEST: Literal[9] = 9

    STORAGE_PLUGINS: Mapping[StoragePlugin, type[StorageWriter]] = {
        StoragePlugin.SQLITE3: Sqlite3Writer,
        StoragePlugin.MCAP: McapWriter,
    }

    def __init__(
        self,
        path: Path | str,
        *,
        version: Literal[8, 9],
        storage_plugin: StoragePlugin = StoragePlugin.SQLITE3,
    ) -> None:
        """Initialize writer.

        Args:
            path: Filesystem path to bag.
            version: Rosbag2 file format version.
            storage_plugin: Storage plugin to use.

        Raises:
            WriterError: Target path exists already, Writer can only create new rosbags.

        """
        path = Path(path)
        if path.exists():
            msg = f'{path} exists already, not overwriting.'
            raise WriterError(msg)

        self.path = path
        self.metapath = path / 'metadata.yaml'
        self.version = version
        self.storage_plugin = storage_plugin

        self.compression_mode = CompressionMode.NONE
        self.compression_format = ''
        self.compressor: zstandard.ZstdCompressor | None = None

        self.storage: StorageWriter | None = None
        self.connections: list[Connection] = []
        self.counts: dict[int, int] = {}
        self.custom_data: dict[str, str] = {}
        self.added_types: set[str] = set()
        self.min_timestamp = 2**63 - 1
        self.max_timestamp = 0

    def set_compression(self, mode: CompressionMode, fmt: CompressionFormat) -> None:
        """Enable compression on bag.

        This function has to be called before opening.

        Args:
            mode: Compression mode to use, either 'file' or 'message'.
            fmt: Compressor to use, currently only 'zstd'.

        Raises:
            WriterError: Bag already open.

        """
        if self.storage:
            msg = f'Cannot set compression, bag {self.path} already open.'
            raise WriterError(msg)
        if mode == CompressionMode.NONE:
            return
        self.compression_mode = mode
        self.compression_format = fmt.name.lower()
        self.compressor = zstandard.ZstdCompressor()

    def set_custom_data(self, key: str, value: str) -> None:
        """Set key value pair in custom_data.

        Args:
            key: Key to set.
            value: Value to set.

        Raises:
            WriterError: If value has incorrect type.

        """
        if not isinstance(value, str):
            msg = f'Cannot set non-string value {value!r} in custom_data.'
            raise WriterError(msg)
        self.custom_data[key] = value

    def open(self) -> None:
        """Open rosbag2 for writing.

        Create base directory and open database connection.

        """
        try:
            self.path.mkdir(mode=0o755, parents=True)
        except FileExistsError:
            msg = f'{self.path} exists already, not overwriting.'
            raise WriterError(msg) from None

        self.storage = self.STORAGE_PLUGINS[self.storage_plugin](self.path, self.compression_mode)

    def add_connection(
        self,
        topic: str,
        msgtype: str,
        *,
        typestore: Typestore | None = None,
        msgdef: str | None = None,
        rihs01: str | None = None,
        serialization_format: str = 'cdr',
        offered_qos_profiles: Sequence[Qos] = (),
    ) -> Connection:
        """Add a connection.

        This function can only be called after opening a bag.

        Args:
            topic: Topic name.
            msgtype: Message type.
            typestore: Typestore.
            msgdef: Message definition.
            rihs01: Message hash.
            serialization_format: Serialization format.
            offered_qos_profiles: QOS Profile.

        Returns:
            Connection object.

        Raises:
            WriterError: Bag not open or topic previously registered.

        """
        if not self.storage:
            msg = 'Bag was not opened.'
            raise WriterError(msg)

        if msgdef is None or rihs01 is None:
            if not typestore:
                msg = (
                    'Cannot determine message definition. '
                    'Use either the typestore or msgdef+rihs01 arguments.'
                )
                raise WriterError(msg)
            msgdef, _ = typestore.generate_msgdef(msgtype, ros_version=2)
            rihs01 = typestore.hash_rihs01(msgtype)
        assert msgdef is not None
        assert rihs01

        fmt = 'ros2idl' if msgdef.startswith('=' * 80 + '\nIDL: ') else 'ros2msg'

        fmtmap = {
            'ros2msg': MessageDefinitionFormat.MSG,
            'ros2idl': MessageDefinitionFormat.IDL,
        }

        connection = Connection(
            id=len(self.connections) + 1,
            topic=topic,
            msgtype=msgtype,
            msgdef=MessageDefinition(fmtmap[fmt], msgdef),
            digest=rihs01,
            msgcount=0,
            ext=ConnectionExtRosbag2(
                serialization_format=serialization_format,
                offered_qos_profiles=list(offered_qos_profiles),
            ),
            owner=self,
        )
        for conn in self.connections:
            if (
                conn.topic == connection.topic
                and conn.msgtype == connection.msgtype
                and conn.ext == connection.ext
            ):
                msg = f'Connection can only be added once: {connection!r}.'
                raise WriterError(msg)

        self.connections.append(connection)
        self.counts[connection.id] = 0

        dump_qos = dump_qos_v9 if self.version >= 9 else dump_qos_v8
        dumped = dump_qos(list(offered_qos_profiles))
        if not isinstance(dumped, str):
            stream = StringIO()
            yaml = YAML(typ='safe')
            yaml.default_flow_style = False
            yaml.dump(dumped, stream)
            dumped = stream.getvalue().strip()

        if msgtype not in self.added_types:
            self.storage.add_msgtype(connection)
            self.added_types.add(msgtype)
        self.storage.add_connection(connection, dumped)
        return connection

    def write(self, connection: Connection, timestamp: int, data: bytes | memoryview) -> None:
        """Write message to rosbag2.

        Args:
            connection: Connection to write message to.
            timestamp: Message timestamp (ns).
            data: Serialized message data.

        Raises:
            WriterError: Bag not open or topic not registered.

        """
        if not self.storage:
            msg = 'Bag was not opened.'
            raise WriterError(msg)
        if connection not in self.connections:
            msg = f'Tried to write to unknown connection {connection!r}.'
            raise WriterError(msg)

        if self.compression_mode == CompressionMode.MESSAGE:
            assert self.compressor
            data = self.compressor.compress(data)

        self.storage.write(connection, timestamp, data)
        self.counts[connection.id] += 1
        self.min_timestamp = min(timestamp, self.min_timestamp)
        self.max_timestamp = max(timestamp, self.max_timestamp)

    def close(self) -> None:
        """Close rosbag2 after writing.

        Closes open database transactions and writes metadata.yaml.

        """
        assert self.storage

        compression_mode = (
            ''
            if self.compression_mode in {CompressionMode.NONE, CompressionMode.STORAGE}
            else self.compression_mode.name.lower()
        )

        compression_format = (
            ''
            if self.compression_mode in {CompressionMode.NONE, CompressionMode.STORAGE}
            else self.compression_format
        )

        path = self.storage.path
        dst = (
            path.with_suffix(f'{path.suffix}.{compression_format}')
            if self.compression_mode == CompressionMode.FILE
            else path
        )

        duration = max(0, self.max_timestamp - self.min_timestamp)
        start = self.min_timestamp
        count = sum(self.counts.values())

        dump_qos = dump_qos_v9 if self.version >= 9 else dump_qos_v8

        metadata: dict[str, Metadata] = {
            'rosbag2_bagfile_information': {
                'version': self.version,
                'storage_identifier': self.storage_plugin.name.lower(),
                'relative_file_paths': [dst.name],
                'duration': {'nanoseconds': duration},
                'starting_time': {'nanoseconds_since_epoch': start},
                'message_count': count,
                'topics_with_message_count': [
                    {
                        'topic_metadata': {
                            'name': x.topic,
                            'type': x.msgtype,
                            'serialization_format': x.ext.serialization_format,
                            'offered_qos_profiles': dump_qos(x.ext.offered_qos_profiles),
                            'type_description_hash': x.digest,
                        },
                        'message_count': self.counts[x.id],
                    }
                    for x in self.connections
                    if isinstance(x.ext, ConnectionExtRosbag2)
                ],
                'compression_format': compression_format,
                'compression_mode': compression_mode,
                'files': [
                    {
                        'path': dst.name,
                        'starting_time': {'nanoseconds_since_epoch': start},
                        'duration': {'nanoseconds': duration},
                        'message_count': count,
                    },
                ],
                'custom_data': self.custom_data or None,
                'ros_distro': 'rosbags',
            },
        }

        yaml = YAML(typ='safe')
        yaml.default_flow_style = False

        metastr = StringIO()
        yaml.dump(metadata, metastr)
        self.metapath.write_text(metastr.getvalue(), 'utf8')

        metastr = StringIO()
        yaml.dump(metadata['rosbag2_bagfile_information'], metastr)
        self.storage.close(self.version, metastr.getvalue().strip())
        self.storage = None

        if self.compression_mode == CompressionMode.FILE:
            assert self.compressor
            with path.open('rb') as infile, dst.open('wb') as outfile:
                _ = self.compressor.copy_stream(infile, outfile)
            path.unlink()

    def __enter__(self) -> Self:
        """Open rosbag2 when entering contextmanager."""
        self.open()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        """Close rosbag2 when exiting contextmanager."""
        self.close()
        return False



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Rosbag2 metadata."""

from __future__ import annotations

from io import StringIO
from typing import TYPE_CHECKING, NamedTuple, TypedDict

from ruamel.yaml import YAML

from rosbags.interfaces import (
    Qos,
    QosDurability,
    QosHistory,
    QosLiveliness,
    QosReliability,
    QosTime,
)

if TYPE_CHECKING:

    class QosTimeDict(TypedDict):
        """Time in seconds and nanoseconds."""

        sec: int
        nsec: int

    class QosDict(TypedDict):
        """QoS parameters."""

        history: str | int
        depth: int
        reliability: str | int
        durability: str | int
        deadline: QosTimeDict
        lifespan: QosTimeDict
        liveliness: str | int
        liveliness_lease_duration: QosTimeDict
        avoid_ros_namespace_conventions: bool


class Duration(TypedDict):
    """Bag duration."""

    nanoseconds: int


class StartingTime(TypedDict):
    """Bag starting time."""

    nanoseconds_since_epoch: int


class TopicMetadata(TypedDict):
    """Topic metadata."""

    name: str
    type: str
    serialization_format: str
    offered_qos_profiles: list[QosDict] | str
    type_description_hash: str


class TopicWithMessageCount(TypedDict):
    """Topic with message count."""

    message_count: int
    topic_metadata: TopicMetadata


class FileInformation(TypedDict):
    """Per file metadata."""

    path: str
    starting_time: StartingTime
    duration: Duration
    message_count: int


class Metadata(TypedDict):
    """Rosbag2 metadata file."""

    version: int
    storage_identifier: str
    relative_file_paths: list[str]
    starting_time: StartingTime
    duration: Duration
    message_count: int
    compression_format: str
    compression_mode: str
    topics_with_message_count: list[TopicWithMessageCount]
    files: list[FileInformation]
    custom_data: dict[str, str] | None
    ros_distro: str


class ReaderMetadata(NamedTuple):
    """Storage Metadata."""

    duration: int
    start_time: int
    end_time: int
    message_count: int
    compression_format: str | None
    compression_mode: str | None
    ros_distro: str | None
    custom: dict[str, str] | None


def parse_qos(dcts: list[QosDict] | str) -> list[Qos]:
    """Parse and normalize QoS parameters."""
    if not dcts:
        return []
    items: list[QosDict] = YAML(typ='safe').load(dcts) if isinstance(dcts, str) else dcts

    res: list[Qos] = []
    for item in items:
        try:
            if isinstance((value := item['history']), int):
                history = QosHistory(value)
            else:
                history = QosHistory.__members__[value.upper()]
        except (KeyError, ValueError):
            history = QosHistory(QosHistory.UNKNOWN)

        try:
            if isinstance((value := item['reliability']), int):
                reliability = QosReliability(value)
            else:
                reliability = QosReliability.__members__[value.upper()]
        except (KeyError, ValueError):
            reliability = QosReliability(QosReliability.UNKNOWN)

        try:
            if isinstance((value := item['durability']), int):
                durability = QosDurability(value)
            else:
                durability = QosDurability.__members__[value.upper()]
        except (KeyError, ValueError):
            durability = QosDurability(QosDurability.UNKNOWN)

        try:
            if isinstance((value := item['liveliness']), int):
                liveliness = QosLiveliness(value)
            else:
                liveliness = QosLiveliness.__members__[value.upper()]
        except (KeyError, ValueError):
            liveliness = QosLiveliness(QosLiveliness.UNKNOWN)

        res.append(
            Qos(
                history,
                item['depth'],
                reliability,
                durability,
                QosTime(item['deadline']['sec'], item['deadline']['nsec']),
                QosTime(item['lifespan']['sec'], item['lifespan']['nsec']),
                liveliness,
                QosTime(
                    item['liveliness_lease_duration']['sec'],
                    item['liveliness_lease_duration']['nsec'],
                ),
                item['avoid_ros_namespace_conventions'],
            ),
        )

    return res


def dump_qos_v8(qos: list[Qos]) -> str:
    """Dump qos for v8 metadata."""
    dcts: list[QosDict] = [
        {
            'history': x.history.value,
            'depth': x.depth,
            'reliability': x.reliability.value,
            'durability': x.durability.value,
            'deadline': {'sec': x.deadline.sec, 'nsec': x.deadline.nsec},
            'lifespan': {'sec': x.lifespan.sec, 'nsec': x.lifespan.nsec},
            'liveliness': x.liveliness.value,
            'liveliness_lease_duration': {
                'sec': x.liveliness_lease_duration.sec,
                'nsec': x.liveliness_lease_duration.nsec,
            },
            'avoid_ros_namespace_conventions': x.avoid_ros_namespace_conventions,
        }
        for x in qos
    ]

    if not dcts:
        return ''

    stream = StringIO()
    YAML(typ='safe').dump(dcts, stream)
    return stream.getvalue().strip()


def dump_qos_v9(qos: list[Qos]) -> list[QosDict]:
    """Dump qos for v9 metadata."""
    return [
        {
            'history': x.history.name.lower(),
            'depth': x.depth,
            'reliability': x.reliability.name.lower(),
            'durability': x.durability.name.lower(),
            'deadline': {'sec': x.deadline.sec, 'nsec': x.deadline.nsec},
            'lifespan': {'sec': x.lifespan.sec, 'nsec': x.lifespan.nsec},
            'liveliness': x.liveliness.name.lower(),
            'liveliness_lease_duration': {
                'sec': x.liveliness_lease_duration.sec,
                'nsec': x.liveliness_lease_duration.nsec,
            },
            'avoid_ros_namespace_conventions': x.avoid_ros_namespace_conventions,
        }
        for x in qos
    ]



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Tool to update builtin types shipped with rosbags."""

from __future__ import annotations

from itertools import groupby
from os import walk
from pathlib import Path
from typing import TYPE_CHECKING

from rosbags.typesys.store import Typestore
from rosbags.typesys.stores import Stores, get_typestore

from .codegen import generate_python_code
from .idl import get_types_from_idl
from .msg import get_types_from_msg

if TYPE_CHECKING:
    from rosbags.interfaces.typing import Typesdict


def generate_api(name: str) -> str:
    """Generate api docs."""
    return '\n'.join(
        [
            f'rosbags.typesys.stores.{name}',
            '=' * (23 + len(name)),
            '',
            f'.. automodule:: rosbags.typesys.stores.{name}',
            '   :members:',
            '   :show-inheritance:',
        ]
    )


def generate_docs(name: str, typs: Typesdict) -> str:
    """Generate types documentation."""
    res = [name, '=' * len(name)]
    for namespace, msgs in groupby([x.split('/msg/') for x in typs], key=lambda x: x[0]):
        res.append(namespace)
        res.append('*' * len(namespace))

        for _, msg in msgs:
            res.append(
                f'- :py:class:`{msg} <rosbags.typesys.stores.{name}.{namespace}__msg__{msg}>`',
            )
        res.append('')
    return '\n'.join(res)


def get_types_from_path(path: Path) -> Typesdict:
    """Walk path and parse all found messages into a typesdict."""
    typs: Typesdict = {}
    for root, dirnames, files in walk(path):
        if '.rosbags_ignore' in files:
            dirnames.clear()
            continue
        for fname in files:
            path = Path(root, fname)
            if path.suffix == '.idl':
                typs.update(get_types_from_idl(path.read_text(encoding='utf-8')))
            elif path.suffix == '.msg':
                name = path.relative_to(path.parents[2]).with_suffix('')
                if '/msg/' not in str(name):
                    name = name.parent / 'msg' / name.name
                typs.update(get_types_from_msg(path.read_text(encoding='utf-8'), str(name)))

    return dict(sorted(typs.items()))


def main(base: str, target: str, pathstr: str) -> None:  # pragma: no cover
    """Update builtin types.

    Discover message definitions in filesystem and generate types.py module.

    """
    selfdir = Path(__file__).parent

    typs = get_types_from_path(Path(pathstr))
    if not base:
        typs = {k: v for k, v in typs.items() if k.startswith('builtin')}

    typestore = get_typestore(Stores(base)) if base else Typestore()
    newstore = Typestore()
    newstore.register(typs)

    remove = sorted(typestore.fielddefs.keys() - newstore.fielddefs.keys())
    add = sorted(newstore.fielddefs.keys() - typestore.fielddefs.keys())
    change = sorted(newstore.fielddefs.keys() & typestore.fielddefs.keys())
    keep: list[str] = []

    for typ in change.copy():
        if typestore.fielddefs[typ] == newstore.fielddefs[typ]:
            keep.append(typ)
            change.remove(typ)

    Typestore().register(typs)
    _ = (
        (selfdir / 'stores' / target)
        .with_suffix('.py')
        .write_text(generate_python_code(typs, base, remove, add, change, keep))
    )
    docsdir = selfdir.parent.parent.parent / 'docs'
    _ = (docsdir / 'api' / 'stores' / f'{target}.rst').write_text(generate_api(target))
    _ = (docsdir / 'topics' / f'typesys-types-{target}.rst').write_text(generate_docs(target, typs))


if __name__ == '__main__':
    import sys

    main(*sys.argv[1:])



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Rosbag Converter Tests."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, cast
from unittest.mock import ANY, MagicMock, call, patch

import pytest

from rosbags.convert import ConverterError, convert
from rosbags.convert.converter import (
    LATCH,
    create_connections_converters,
    default_message,
    generate_message_converter,
    migrate_bytes,
    migrate_message,
)
from rosbags.highlevel import AnyReaderError
from rosbags.interfaces import (
    Connection,
    ConnectionExtRosbag1,
    ConnectionExtRosbag2,
    MessageDefinition,
    MessageDefinitionFormat,
    Qos,
)
from rosbags.rosbag1 import (
    Writer as Writer1,
    WriterError as WriterError1,
)
from rosbags.rosbag2 import (
    Writer as Writer2,
    WriterError as WriterError2,
)
from rosbags.typesys.msg import get_types_from_msg
from rosbags.typesys.stores import Stores, get_typestore

if TYPE_CHECKING:
    from pathlib import Path

    from rosbags.typesys.stores.ros1_noetic import (
        sensor_msgs__msg__CameraInfo as CameraInfo1,
        shape_msgs__msg__MeshTriangle as MeshTriangle1,
        std_msgs__msg__Duration as Duration1,
        std_msgs__msg__Int8 as Int8,
        std_msgs__msg__Int8MultiArray as Int8MultiArray1,
        visualization_msgs__msg__InteractiveMarkerUpdate as InteractiveMarkerUpdate1,
        visualization_msgs__msg__Marker as Marker1,
    )
    from rosbags.typesys.stores.ros2_iron import (
        sensor_msgs__msg__CameraInfo as CameraInfo2,
        shape_msgs__msg__MeshTriangle as MeshTriangle2,
        std_msgs__msg__Int16MultiArray as Int16MultiArray2,
        std_msgs__msg__String as String2,
        visualization_msgs__msg__InteractiveMarkerUpdate as InteractiveMarkerUpdate2,
        visualization_msgs__msg__Marker as Marker2,
    )


def test_convert_reader_errors(tmp_path: Path) -> None:
    """Test convert forwards reader errors."""
    with (
        patch('rosbags.convert.converter.AnyReader', side_effect=AnyReaderError('exc')),
        pytest.raises(ConverterError, match='Reading source bag: exc'),
    ):
        convert([], tmp_path / 'foo', 'sqlite3', 8, None, 'file', None, None, (), (), (), ())


def test_convert_writer_errors(tmp_path: Path) -> None:
    """Test convert forwards writer errors."""
    with (
        patch('rosbags.convert.converter.Writer2', side_effect=WriterError2('exc')),
        pytest.raises(ConverterError, match='Writing destination bag: exc'),
    ):
        convert([], tmp_path / 'foo', 'sqlite3', 8, None, 'file', None, None, (), (), (), ())

    with (
        patch('rosbags.convert.converter.Writer1', side_effect=WriterError1('exc')),
        pytest.raises(ConverterError, match='Writing destination bag: exc'),
    ):
        convert([], tmp_path / 'foo.bag', 'sqlite3', 8, None, 'file', None, None, (), (), (), ())


def test_convert_reraises_assertions(tmp_path: Path) -> None:
    """Test convert reraises assertion errors."""
    with (
        patch('rosbags.convert.converter.AnyReader', side_effect=AssertionError('exc')),
        pytest.raises(AssertionError, match='exc'),
    ):
        convert([], tmp_path / 'foo', 'sqlite3', 8, None, 'file', None, None, (), (), (), ())


def test_convert_forwards_exceptions(tmp_path: Path) -> None:
    """Test convert forwards exceptions."""
    with (
        patch('rosbags.convert.converter.AnyReader', side_effect=KeyError('exc')),
        pytest.raises(ConverterError, match="Converting rosbag: KeyError\\('exc'\\)"),
    ):
        convert([], tmp_path / 'foo', 'sqlite3', 8, None, 'file', None, None, (), (), (), ())


def test_convert_enables_compression(tmp_path: Path) -> None:
    """Test convert forwards exceptions."""
    with (
        patch('rosbags.convert.converter.AnyReader'),
        patch('rosbags.convert.converter.Writer1') as writer,
    ):
        convert([], tmp_path / 'foo.bag', 'sqlite3', 8, None, 'file', None, None, (), (), (), ())
    writer.return_value.set_compression.assert_not_called()

    with (
        patch('rosbags.convert.converter.AnyReader'),
        patch('rosbags.convert.converter.Writer1') as writer,
    ):
        convert([], tmp_path / 'foo.bag', 'sqlite3', 8, 'bz2', 'file', None, None, (), (), (), ())
    writer.return_value.set_compression.assert_called()

    with (
        patch('rosbags.convert.converter.AnyReader'),
        patch('rosbags.convert.converter.Writer2') as writer,
    ):
        convert([], tmp_path / 'foo', 'sqlite3', 8, None, 'file', None, None, (), (), (), ())
    writer.return_value.set_compression.assert_not_called()

    with (
        patch('rosbags.convert.converter.AnyReader'),
        patch('rosbags.convert.converter.Writer2') as writer,
    ):
        convert([], tmp_path / 'foo', 'sqlite3', 8, 'zstd', 'file', None, None, (), (), (), ())
    writer.return_value.set_compression.assert_called()


def test_convert_connection_filtering(tmp_path: Path) -> None:
    """Test convert filters connections."""
    with (
        patch('rosbags.convert.converter.AnyReader') as reader,
        patch('rosbags.convert.converter.Writer2'),
        patch('rosbags.convert.converter.create_connections_converters') as ccc,
    ):
        reader.return_value.__enter__.return_value.connections = []
        convert([], tmp_path, 'sqlite3', 8, None, 'file', None, None, (), (), (), ())
    ccc.assert_not_called()

    with (
        patch('rosbags.convert.converter.AnyReader') as reader,
        patch('rosbags.convert.converter.Writer2'),
        patch(
            'rosbags.convert.converter.create_connections_converters',
            side_effect=AnyReaderError,
        ) as ccc,
    ):
        conn = MagicMock()
        conn.topic = 'foo'
        conn.msgtype = 'bar'
        reader.return_value.__enter__.return_value.connections = [conn]
        with pytest.raises(ConverterError):
            convert([], tmp_path, 'sqlite3', 8, None, 'file', None, None, (), (), (), ())
        ccc.reset_mock()

        with pytest.raises(ConverterError):
            convert(
                [], tmp_path, 'sqlite3', 8, None, 'file', None, None, (), ('foo'), (), ('unknown')
            )
        ccc.reset_mock()

        with pytest.raises(ConverterError):
            convert(
                [], tmp_path, 'sqlite3', 8, None, 'file', None, None, (), ('unknown'), (), ('bar')
            )
        ccc.reset_mock()

        convert([], tmp_path, 'sqlite3', 8, None, 'file', None, None, ('foo'), (), (), ())
        ccc.assert_not_called()

        convert([], tmp_path, 'sqlite3', 8, None, 'file', None, None, (), (), ('bar'), ())
        ccc.assert_not_called()

        convert([], tmp_path, 'sqlite3', 8, None, 'file', None, None, (), ('unknown'), (), ())
        ccc.assert_not_called()

        convert([], tmp_path, 'sqlite3', 8, None, 'file', None, None, (), (), (), ('unknown'))
        ccc.assert_not_called()


def test_convert_applies_transforms(tmp_path: Path) -> None:
    """Test convert applies transforms."""
    with (
        patch('rosbags.convert.converter.AnyReader') as reader,
        patch('rosbags.convert.converter.Writer2') as writer,
        patch(
            'rosbags.convert.converter.create_connections_converters',
        ) as ccc,
    ):
        conn = MagicMock()
        conn.id = 42
        conn.topic = 'foo'
        conn.msgtype = 'bar'
        conn.owner = 'own'
        rctx = reader.return_value.__enter__.return_value
        rctx.connections = [conn]
        rctx.messages.return_value = [(conn, 1, 2)]
        ccc.return_value = [
            {(42, 'own'): 666},
            {'bar': lambda x: x * 2},
        ]

        convert([], tmp_path, 'sqlite3', 8, None, 'file', None, None, (), (), (), ())

        writer.return_value.write.assert_called_with(666, 1, 4)


def test_connection_params_are_converted() -> None:
    """Test connection params are converted."""
    connections = [
        Connection(
            1,
            '/t1',
            'std_msgs/msg/Int8',
            MessageDefinition(MessageDefinitionFormat.NONE, ''),
            '',
            0,
            ConnectionExtRosbag1(None, latching=1),
            None,
        ),
        Connection(
            2,
            '/t2',
            'std_msgs/msg/Int8',
            MessageDefinition(MessageDefinitionFormat.NONE, ''),
            '',
            0,
            ConnectionExtRosbag1('caller', latching=0),
            None,
        ),
    ]

    reader = MagicMock()
    reader.is2 = False
    reader.typestore = get_typestore(Stores.ROS1_NOETIC)

    writer = MagicMock(spec=Writer1)
    writer.connections = []

    _ = create_connections_converters(connections, None, reader, writer)
    writer.add_connection.assert_has_calls(
        [
            call(
                '/t1',
                'std_msgs/msg/Int8',
                typestore=ANY,
                callerid=None,
                latching=1,
            ),
            call(
                '/t2',
                'std_msgs/msg/Int8',
                typestore=ANY,
                callerid='caller',
                latching=0,
            ),
        ]
    )

    writer = MagicMock(spec=Writer2)
    writer.connections = []
    _ = create_connections_converters(connections, None, reader, writer)
    writer.add_connection.assert_has_calls(
        [
            call(
                '/t1',
                'std_msgs/msg/Int8',
                typestore=ANY,
                serialization_format='cdr',
                offered_qos_profiles=LATCH,
            ),
            call(
                '/t2',
                'std_msgs/msg/Int8',
                typestore=ANY,
                serialization_format='cdr',
                offered_qos_profiles=[],
            ),
        ]
    )

    connections = [
        Connection(
            1,
            '/t1',
            'std_msgs/msg/Int8',
            MessageDefinition(MessageDefinitionFormat.NONE, ''),
            '',
            0,
            ConnectionExtRosbag2('cdr', LATCH),
            None,
        ),
        Connection(
            2,
            '/t2',
            'std_msgs/msg/Int8',
            MessageDefinition(MessageDefinitionFormat.NONE, ''),
            '',
            0,
            ConnectionExtRosbag2('cdr', []),
            None,
        ),
    ]

    reader = MagicMock()
    reader.is2 = True
    reader.typestore = get_typestore(Stores.ROS2_FOXY)

    writer = MagicMock(spec=Writer1)
    writer.connections = []

    _ = create_connections_converters(connections, None, reader, writer)
    writer.add_connection.assert_has_calls(
        [
            call(
                '/t1',
                'std_msgs/msg/Int8',
                typestore=ANY,
                callerid=None,
                latching=1,
            ),
            call(
                '/t2',
                'std_msgs/msg/Int8',
                typestore=ANY,
                callerid=None,
                latching=0,
            ),
        ]
    )

    writer = MagicMock(spec=Writer2)
    writer.connections = []
    _ = create_connections_converters(connections, None, reader, writer)
    writer.add_connection.assert_has_calls(
        [
            call(
                '/t1',
                'std_msgs/msg/Int8',
                typestore=ANY,
                serialization_format='cdr',
                offered_qos_profiles=LATCH,
            ),
            call(
                '/t2',
                'std_msgs/msg/Int8',
                typestore=ANY,
                serialization_format='cdr',
                offered_qos_profiles=[],
            ),
        ]
    )


def test_destination_typestore_gets_created() -> None:
    """Test destination typestore gets created."""
    connections = [
        Connection(
            1,
            '/t1',
            'std_msgs/msg/Int8',
            MessageDefinition(MessageDefinitionFormat.NONE, ''),
            '',
            0,
            ConnectionExtRosbag1('', latching=True),
            None,
        ),
    ]

    typestore1 = get_typestore(Stores.ROS1_NOETIC)
    typestore2 = get_typestore(Stores.ROS2_FOXY)

    reader = MagicMock()
    reader.is2 = False
    reader.typestore = typestore1

    writer = MagicMock(spec=Writer1)
    writer.connections = []

    _ = create_connections_converters(connections, None, reader, writer)
    writer.add_connection.assert_called_with(
        '/t1',
        'std_msgs/msg/Int8',
        typestore=typestore1,
        callerid='',
        latching=True,
    )

    reader = MagicMock()
    reader.is2 = True
    reader.typestore = typestore2

    writer = MagicMock(spec=Writer1)
    writer.connections = []

    _ = create_connections_converters(connections, None, reader, writer)
    writer.add_connection.assert_called_with(
        '/t1',
        'std_msgs/msg/Int8',
        typestore=ANY,
        callerid='',
        latching=True,
    )
    assert writer.add_connection.call_args.kwargs['typestore'] != typestore2

    reader = MagicMock()
    reader.is2 = False
    reader.typestore = typestore1

    writer = MagicMock(spec=Writer2)
    writer.connections = []

    _ = create_connections_converters(connections, None, reader, writer)
    writer.add_connection.assert_called_with(
        '/t1',
        'std_msgs/msg/Int8',
        typestore=ANY,
        serialization_format='cdr',
        offered_qos_profiles=LATCH,
    )
    assert writer.add_connection.call_args.kwargs['typestore'] != typestore1

    reader = MagicMock()
    reader.is2 = True
    reader.typestore = typestore2

    writer = MagicMock(spec=Writer2)
    writer.connections = []

    _ = create_connections_converters(connections, None, reader, writer)
    writer.add_connection.assert_called_with(
        '/t1',
        'std_msgs/msg/Int8',
        typestore=typestore2,
        serialization_format='cdr',
        offered_qos_profiles=LATCH,
    )

    reader = MagicMock()
    reader.is2 = False
    reader.typestore = typestore1

    writer = MagicMock(spec=Writer1)
    writer.connections = []

    empty = get_typestore(Stores.EMPTY)
    _ = create_connections_converters(connections, empty, reader, writer)
    writer.add_connection.assert_called_with(
        '/t1',
        'std_msgs/msg/Int8',
        typestore=empty,
        callerid='',
        latching=True,
    )
    assert 'std_msgs/msg/Int8' in empty.types


def test_connection_deduplication() -> None:
    """Test connections get deduplicated."""
    connections = [
        Connection(
            1,
            '/t1',
            'std_msgs/msg/Int8',
            MessageDefinition(MessageDefinitionFormat.NONE, ''),
            '',
            0,
            ConnectionExtRosbag1('', latching=True),
            None,
        ),
        Connection(
            2,
            '/t1',
            'std_msgs/msg/Int8',
            MessageDefinition(MessageDefinitionFormat.NONE, ''),
            '',
            0,
            ConnectionExtRosbag1('', latching=True),
            None,
        ),
        Connection(
            3,
            '/t1',
            'std_msgs/msg/Int8',
            MessageDefinition(MessageDefinitionFormat.NONE, ''),
            '',
            0,
            ConnectionExtRosbag1('', latching=False),
            None,
        ),
        Connection(
            4,
            '/t1',
            'std_msgs/msg/Int8',
            MessageDefinition(MessageDefinitionFormat.NONE, ''),
            '',
            0,
            ConnectionExtRosbag1('foo', latching=False),
            None,
        ),
    ]

    typestore = get_typestore(Stores.ROS1_NOETIC)
    reader = MagicMock()
    reader.is2 = False
    reader.typestore = typestore

    writer = MagicMock(spec=Writer1)
    writer.connections = []

    def add_connection1(
        name: str, typ: str, *, typestore: object, callerid: str, latching: bool
    ) -> Connection:
        _ = typestore
        res = Connection(
            len(writer.connections),
            name,
            typ,
            MessageDefinition(MessageDefinitionFormat.NONE, ''),
            '',
            0,
            ConnectionExtRosbag1(callerid, latching),
            None,
        )
        writer.connections.append(res)
        return res

    writer.add_connection = add_connection1

    res = create_connections_converters(connections, typestore, reader, writer)
    assert res[0] == {
        (1, None): writer.connections[0],
        (2, None): writer.connections[0],
        (3, None): writer.connections[1],
        (4, None): writer.connections[2],
    }
    assert len(res[1]) == 1

    reader = MagicMock()
    reader.is2 = False
    reader.typestore = typestore

    writer = MagicMock(spec=Writer2)
    writer.connections = []

    def add_connection2(
        name: str,
        typ: str,
        *,
        typestore: object,
        serialization_format: str,
        offered_qos_profiles: list[Qos],
    ) -> Connection:
        _ = typestore
        res = Connection(
            len(writer.connections),
            name,
            typ,
            MessageDefinition(MessageDefinitionFormat.NONE, ''),
            '',
            0,
            ConnectionExtRosbag2(serialization_format, offered_qos_profiles),
            None,
        )
        writer.connections.append(res)
        return res

    writer.add_connection = add_connection2

    res = create_connections_converters(connections, typestore, reader, writer)
    assert res[0] == {
        (1, None): writer.connections[0],
        (2, None): writer.connections[0],
        (3, None): writer.connections[1],
        (4, None): writer.connections[1],
    }
    assert len(res[1]) == 1


def test_message_converter() -> None:
    """Test message converter."""
    srcts = get_typestore(Stores.ROS1_NOETIC)
    dstts = get_typestore(Stores.ROS2_IRON)

    conv = generate_message_converter(
        srcts,
        dstts,
        'builtin_interfaces/msg/Time',
        'builtin_interfaces/msg/Time',
        {},
        src_is2=False,
        dst_is2=False,
    )
    assert conv(b'42').tobytes() == b'42'

    conv = generate_message_converter(
        srcts,
        dstts,
        'builtin_interfaces/msg/Time',
        'builtin_interfaces/msg/Time',
        {},
        src_is2=True,
        dst_is2=True,
    )
    assert conv(b'42').tobytes() == b'42'

    conv = generate_message_converter(
        srcts,
        dstts,
        'std_msgs/msg/Header',
        'std_msgs/msg/Header',
        {},
        src_is2=False,
        dst_is2=False,
    )
    assert conv(b'42').tobytes() == b'42'

    conv = generate_message_converter(
        srcts,
        srcts,  # same typestore
        'visualization_msgs/msg/MarkerArray',
        'visualization_msgs/msg/MarkerArray',
        {},
        src_is2=False,
        dst_is2=False,
    )
    assert conv(b'42').tobytes() == b'42'

    with patch.object(dstts, 'ros1_to_cdr', return_value=memoryview(b'mock')) as func:
        conv = generate_message_converter(
            srcts,
            dstts,
            'builtin_interfaces/msg/Time',
            'builtin_interfaces/msg/Time',
            {},
            src_is2=False,
            dst_is2=True,
        )
        assert conv(b'42').tobytes() == b'mock'
        func.assert_called_with(b'42', typename='builtin_interfaces/msg/Time')

    with patch.object(srcts, 'cdr_to_ros1', return_value=memoryview(b'mock')) as func:
        conv = generate_message_converter(
            srcts,
            dstts,
            'builtin_interfaces/msg/Time',
            'builtin_interfaces/msg/Time',
            {},
            src_is2=True,
            dst_is2=False,
        )
        assert conv(b'42').tobytes() == b'mock'
        func.assert_called_with(b'42', typename='builtin_interfaces/msg/Time')

    with patch('rosbags.convert.converter.migrate_bytes', return_value=memoryview(b'mock')) as func:
        conv = generate_message_converter(
            srcts,
            dstts,
            'visualization_msgs/msg/MarkerArray',
            'visualization_msgs/msg/MarkerArray',
            {},
            src_is2=False,
            dst_is2=False,
        )
        assert conv(b'42').tobytes() == b'mock'
        func.assert_called()

    with patch('rosbags.convert.converter.migrate_bytes', return_value=memoryview(b'mock')) as func:
        conv = generate_message_converter(
            srcts,
            dstts,
            'std_msgs/msg/Int8',
            'std_msgs/msg/Int16',
            {},
            src_is2=False,
            dst_is2=False,
        )
        assert conv(b'42').tobytes() == b'mock'
        func.assert_called()


def test_migrate_bytes() -> None:
    """Test byte level conversion."""
    srcts = get_typestore(Stores.ROS1_NOETIC)
    dstts = get_typestore(Stores.ROS2_IRON)

    res = migrate_bytes(
        srcts,
        dstts,
        'std_msgs/msg/Int8',
        'std_msgs/msg/Int8',
        {},
        b'\x01',
        src_is2=False,
        dst_is2=False,
    )
    assert res.tobytes() == b'\x01'

    res = migrate_bytes(
        srcts,
        dstts,
        'std_msgs/msg/Int8',
        'std_msgs/msg/Int8',
        {},
        res,
        src_is2=False,
        dst_is2=True,
    )
    assert res.tobytes() == b'\x00\x01\x00\x00\x01'

    res = migrate_bytes(
        srcts,
        dstts,
        'std_msgs/msg/Int8',
        'std_msgs/msg/Int8',
        {},
        res,
        src_is2=True,
        dst_is2=True,
    )
    assert res.tobytes() == b'\x00\x01\x00\x00\x01'

    res = migrate_bytes(
        srcts,
        dstts,
        'std_msgs/msg/Int8',
        'std_msgs/msg/Int8',
        {},
        res,
        src_is2=True,
        dst_is2=False,
    )
    assert res.tobytes() == b'\x01'


def test_migrate_message() -> None:
    """Test message level conversion."""
    srcts = get_typestore(Stores.ROS1_NOETIC)
    dstts = get_typestore(Stores.ROS2_IRON)
    cache1: dict[str, object] = {}

    msg = cast('Marker1', default_message(srcts, 'visualization_msgs/msg/Marker'))
    res = cast(
        'Marker2',
        migrate_message(
            srcts,
            dstts,
            'visualization_msgs/msg/Marker',
            'visualization_msgs/msg/Marker',
            cache1,
            msg,
        ),
    )
    assert res.texture_resource == ''
    assert isinstance(res.texture, dstts.types['sensor_msgs/msg/CompressedImage'])
    assert res.uv_coordinates == []
    assert isinstance(res.mesh_file, dstts.types['visualization_msgs/msg/MeshFile'])

    cache2: dict[str, object] = {}
    assert (
        migrate_message(
            dstts,
            srcts,
            'visualization_msgs/msg/Marker',
            'visualization_msgs/msg/Marker',
            cache2,
            res,
        )
        == msg
    )

    msg2 = cast(
        'InteractiveMarkerUpdate1',
        default_message(
            srcts,
            'visualization_msgs/msg/InteractiveMarkerUpdate',
        ),
    )
    msg2.erases.append('foo')
    res2 = cast(
        'InteractiveMarkerUpdate2',
        migrate_message(
            srcts,
            dstts,
            'visualization_msgs/msg/InteractiveMarkerUpdate',
            'visualization_msgs/msg/InteractiveMarkerUpdate',
            cache1,
            msg2,
        ),
    )
    assert res2.erases == ['foo']

    # Type change.
    msg3 = cast(
        'Int8',
        default_message(
            srcts,
            'std_msgs/msg/Int8',
        ),
    )
    msg3.data = 42
    res3 = cast(
        'String2',
        migrate_message(
            srcts,
            dstts,
            'std_msgs/msg/Int8',
            'std_msgs/msg/String',
            cache1,
            msg3,
        ),
    )
    assert res3.data == ''

    # Dtype change.
    msg4 = cast(
        'Int8MultiArray1',
        default_message(
            srcts,
            'std_msgs/msg/Int8MultiArray',
        ),
    )
    msg4.data.resize(4, refcheck=False)
    res4 = cast(
        'Int16MultiArray2',
        migrate_message(
            srcts,
            dstts,
            'std_msgs/msg/Int8MultiArray',
            'std_msgs/msg/Int16MultiArray',
            cache1,
            msg4,
        ),
    )
    assert len(res4.data) == 4
    assert res4.data.dtype.name == 'int16'

    msg5 = cast(
        'Int8',
        default_message(
            srcts,
            'std_msgs/msg/Int8',
        ),
    )
    msg5.data = 42
    res5 = cast(
        'Duration1',
        migrate_message(
            dstts,
            srcts,
            'std_msgs/msg/Int8',
            'std_msgs/msg/Duration',
            cache2,
            msg5,
        ),
    )
    assert res5.data.sec == 0

    msg6 = cast(
        'CameraInfo1',
        default_message(
            srcts,
            'sensor_msgs/msg/CameraInfo',
        ),
    )
    msg6.K[:] = 10
    res6 = cast(
        'CameraInfo2',
        migrate_message(
            srcts,
            dstts,
            'sensor_msgs/msg/CameraInfo',
            'sensor_msgs/msg/CameraInfo',
            cache1,
            msg6,
        ),
    )
    assert (res6.k == 10).all()

    msg7 = cast(
        'MeshTriangle1',
        default_message(
            srcts,
            'shape_msgs/msg/MeshTriangle',
        ),
    )
    msg7.vertex_indices.resize(10, refcheck=False)
    res7 = cast(
        'MeshTriangle2',
        migrate_message(
            srcts,
            dstts,
            'shape_msgs/msg/MeshTriangle',
            'shape_msgs/msg/MeshTriangle',
            cache1,
            msg7,
        ),
    )
    assert len(res7.vertex_indices) == 3

    msg7 = cast(
        'MeshTriangle1',
        default_message(
            srcts,
            'shape_msgs/msg/MeshTriangle',
        ),
    )
    msg7.vertex_indices.resize(1, refcheck=False)
    res7 = cast(
        'MeshTriangle2',
        migrate_message(
            srcts,
            dstts,
            'shape_msgs/msg/MeshTriangle',
            'shape_msgs/msg/MeshTriangle',
            cache1,
            msg7,
        ),
    )
    assert len(res7.vertex_indices) == 3

    @dataclass
    class Strarr:
        data: list[Int8]

    @dataclass
    class Addrem:
        value: int

    typs = get_types_from_msg('string[4] data', 'x/msg/strarr')
    srcts.register(typs)
    typs = get_types_from_msg('string[2] data', 'x/msg/strarr')
    dstts.register(typs)
    msg8 = cast(
        'Strarr',
        default_message(
            srcts,
            'x/msg/strarr',
        ),
    )
    assert len(msg8.data) == 4
    res8 = cast(
        'Strarr',
        migrate_message(
            srcts,
            dstts,
            'x/msg/strarr',
            'x/msg/strarr',
            cache1,
            msg8,
        ),
    )
    assert len(res8.data) == 2

    msg8 = cast(
        'Strarr',
        default_message(
            dstts,
            'x/msg/strarr',
        ),
    )
    assert len(msg8.data) == 2
    res8 = cast(
        'Strarr',
        migrate_message(
            dstts,
            srcts,
            'x/msg/strarr',
            'x/msg/strarr',
            cache1,
            msg8,
        ),
    )
    assert len(res8.data) == 4

    typs = get_types_from_msg('std_msgs/msg/Int8[2] data', 'x/msg/msgarr')
    dstts.register(typs)
    msg9 = cast(
        'Strarr',
        default_message(
            srcts,
            'x/msg/strarr',
        ),
    )
    assert len(msg9.data) == 4
    res9 = cast(
        'Strarr',
        migrate_message(
            srcts,
            dstts,
            'x/msg/strarr',
            'x/msg/msgarr',
            cache1,
            msg9,
        ),
    )
    msg5 = cast(
        'Int8',
        default_message(
            dstts,
            'std_msgs/msg/Int8',
        ),
    )
    assert res9.data == [msg5, msg5]

    typs = get_types_from_msg('uint8 value', 'x/msg/addrem')
    dstts.register(typs)
    msg10 = cast(
        'Int8',
        default_message(
            srcts,
            'std_msgs/msg/Int8',
        ),
    )
    msg10.data = 42
    res10 = cast(
        'Addrem',
        migrate_message(
            srcts,
            dstts,
            'std_msgs/msg/Int8',
            'x/msg/addrem',
            cache1,
            msg10,
        ),
    )
    assert res10.value == 0



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Code generators for the extensible type system."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, cast

from rosbags.interfaces import Nodetype

if TYPE_CHECKING:
    from collections.abc import Sequence

    from rosbags.interfaces.typing import Basename, FieldDesc, Typesdict


INTLIKE = re.compile('^u?(bool|int|float)')


def get_typehint(desc: FieldDesc) -> str:
    """Get python type hint for field.

    Args:
        desc: Field descriptor.

    Returns:
        Type hint for field.

    """
    if desc[0] == Nodetype.BASE:
        if desc[1][0] == 'string':
            return 'str'
        typ = cast(
            'Basename',
            (
                'int'
                if desc[1][0].split('unsigned ')[-1]
                in {'char', 'byte', 'octet', 'short', 'long', 'long long'}
                else desc[1][0]
            ),
        )
        match = INTLIKE.match(typ)
        assert match, typ
        return match.group(1)

    if desc[0] == Nodetype.NAME:
        assert isinstance(desc[1], str)
        return desc[1].replace('/', '__')

    assert desc[0] in {Nodetype.ARRAY, Nodetype.SEQUENCE}
    sub = desc[1][0]
    if sub[0] == Nodetype.BASE:
        typ = sub[1][0]
        if typ.split('unsigned ')[-1] in {
            'byte',
            'char',
            'octet',
            'short',
            'long',
            'long long',
        } or INTLIKE.match(typ):
            typ = cast(
                'Basename',
                {
                    'bool': 'bool_',
                    'byte': 'uint8',
                    'char': 'uint8',
                    'octet': 'uint8',
                    'short': 'int16',
                    'long': 'int32',
                    'long long': 'int64',
                    'unsigned short': 'uint16',
                    'unsigned long': 'uint32',
                    'unsigned long long': 'uint64',
                }.get(typ, typ),
            )
            return f'np.ndarray[tuple[int, ...], np.dtype[np.{typ}]]'

    return f'list[{get_typehint(sub)}]'


def generate_python_code(
    typs: Typesdict,
    base: str | None = None,
    remove: Sequence[str] = (),
    add: Sequence[str] = (),
    change: Sequence[str] = (),
    keep: Sequence[str] = (),
) -> str:
    """Generate python code from types dictionary.

    Args:
        typs: Dictionary mapping message typenames to parsetrees.
        base: Base type store.
        remove: Types to remove.
        add: Types to add.
        change: Types to change.
        keep: Types to keep.

    Returns:
        Code for importable python module.

    """
    _ = remove
    if not base:
        add = list(typs.keys())
    lines = [
        '# Copyright 2020-2026 Ternaris',
        '# SPDX-License-Identifier'  # avoid reuse trigger
        ': Apache-2.0',
        '#',
        '# THIS FILE IS GENERATED, DO NOT EDIT',
        '"""Message type definitions."""',
        '',
        '# ruff: noqa: N801,N816,RUF100,TC004',
        '',
        'from __future__ import annotations',
        '',
    ]
    if add or change:
        lines += ['from dataclasses import dataclass']
    lines += [
        'from typing import TYPE_CHECKING',
        '',
    ]
    if add or change:
        lines += [
            'from rosbags.interfaces import Nodetype as T',
            '',
        ]
    if base:
        lines += [
            f'from . import {base} as base',
            '',
        ]

    lines += [
        'if TYPE_CHECKING:',
    ]
    if add or change:
        lines += [
            '    from typing import ClassVar',
            '',
        ]
    lines += [
        '    import numpy as np',
        '',
        '    from rosbags.interfaces.typing import Typesdict',
        '',
    ]
    lines += ['']

    if keep:
        for name in keep:
            pyname = name.replace('/', '__')
            lines += [f'{pyname} = base.{pyname}']
        lines += ['', '']

    for name, (consts, fields) in typs.items():
        if name not in add and name not in change:
            continue
        pyname = name.replace('/', '__')
        lines += [
            '@dataclass',
            f'class {pyname}:',
            f'    """Class for {name}."""',
            '',
            *[
                (
                    f'    {fname}: {get_typehint(desc)}'
                    f'{" = 0" if fname == "structure_needs_at_least_one_member" else ""}'
                )
                for fname, desc in fields
                or [('structure_needs_at_least_one_member', (Nodetype.BASE, ('uint8', 0)))]
            ],
            *[
                f'    {fname}: ClassVar[{get_typehint((Nodetype.BASE, (ftype, 0)))}] = {fvalue!r}'
                for fname, ftype, fvalue in consts
            ],
            f'    __msgtype__: ClassVar[str] = {name!r}',
        ]

        lines += [
            '',
            '',
        ]

    def get_ftype(ftype: FieldDesc) -> str:
        typs = ['', 'T.BASE', 'T.NAME', 'T.ARRAY', 'T.SEQUENCE']
        if ftype[0] == Nodetype.BASE or ftype[0] == Nodetype.NAME:
            return f'({typs[ftype[0]]}, {ftype[1]!r})'
        return f'({typs[ftype[0]]}, (({typs[ftype[1][0][0]]}, {ftype[1][0][1]!r}), {ftype[1][1]}))'

    if base:
        lines += [
            'FIELDDEFS: Typesdict = {',
            # '    **base.FIELDDEFS,',
        ]
        for typ in keep:
            lines += [f'    {typ!r}: base.FIELDDEFS[{typ!r}],']
    else:
        lines += ['FIELDDEFS: Typesdict = {']
    for name, (consts, fields) in typs.items():
        if name not in add and name not in change:
            continue
        pyname = name.replace('/', '__')
        lines += [
            f"    '{name}': (",
            *(
                [
                    '        [',
                    *[
                        f'            ({fname!r}, {ftype!r}, {fvalue!r}),'
                        for fname, ftype, fvalue in consts
                    ],
                    '        ],',
                ]
                if consts
                else ['        [],']
            ),
            '        [',
            *[
                f'            ({fname!r}, {get_ftype(ftype)}),'
                for fname, ftype in fields
                or [('structure_needs_at_least_one_member', (Nodetype.BASE, ('uint8', 0)))]
            ],
            '        ],',
            '    ),',
        ]
    lines += [
        '}',
        '',
    ]
    if not any('np.ndarray' in x for x in lines):
        idx = lines.index('    import numpy as np')
        _ = lines.pop(idx)
        _ = lines.pop(idx)
    return '\n'.join(lines)



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Types and helpers used by message definition converters."""

from __future__ import annotations

import keyword
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from rosbags.interfaces.typing import Typesdict

    from .peg import Visitor


class TypesysError(Exception):
    """Parser error."""


def normalize_fieldname(name: str) -> str:
    """Normalize field name.

    Avoid collisions with Python keywords.

    Args:
        name: Field name.

    Returns:
        Normalized name.

    """
    if keyword.iskeyword(name):
        return f'{name}_'
    return name


def parse_message_definition(visitor: Visitor, text: str) -> Typesdict:
    """Parse message definition.

    Args:
        visitor: Visitor instance to use.
        text: Message definition.

    Returns:
        Parsetree of message.

    Raises:
        TypesysError: Message parsing failed.

    """
    try:
        rule = visitor.RULES['specification']
        pos = rule.skip_ws(text, 0)
        npos, trees = rule.parse(text, pos)
        assert npos == len(text), f'Could not parse: {text!r}'
        return visitor.visit(trees)  # type: ignore[return-value]
    except Exception as err:
        msg = f'Could not parse: {text!r}'
        raise TypesysError(msg) from err



# Copyright 2020-2026 Ternaris
# SPDX-License-Identifier: Apache-2.0
"""Writer Tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from rosbags.interfaces import (
    Connection,
    ConnectionExtRosbag2,
    MessageDefinition,
    MessageDefinitionFormat,
)
from rosbags.rosbag2 import CompressionFormat, CompressionMode, Writer, WriterError
from rosbags.typesys import Stores, get_typestore

if TYPE_CHECKING:
    from pathlib import Path


def test_writer_writes_storage_and_metadata(tmp_path: Path) -> None:
    """Test writer writes storage file and metadata."""
    store = get_typestore(Stores.LATEST)
    path = tmp_path / 'rosbag2'
    with Writer(path, version=Writer.VERSION_LATEST) as bag:
        connection = bag.add_connection('/test', 'std_msgs/msg/Int8', typestore=store)
        bag.write(connection, 42, b'\x00')
        bag.write(connection, 666, b'\x01' * 4096)
    assert (path / 'metadata.yaml').exists()
    assert 'RIHS01_' in (path / 'metadata.yaml').read_text()
    assert (path / 'rosbag2.db3').exists()
    size = (path / 'rosbag2.db3').stat().st_size

    path = tmp_path / 'compress_none'
    bag = Writer(path, version=Writer.VERSION_LATEST)
    bag.set_compression(CompressionMode.NONE, CompressionFormat.ZSTD)
    with bag:
        connection = bag.add_connection('/test', 'std_msgs/msg/Int8', typestore=store)
        bag.write(connection, 42, b'\x00')
        bag.write(connection, 666, b'\x01' * 4096)
    assert (path / 'metadata.yaml').exists()
    assert (path / 'compress_none.db3').exists()
    assert size == (path / 'compress_none.db3').stat().st_size

    path = tmp_path / 'compress_file'
    bag = Writer(path, version=Writer.VERSION_LATEST)
    bag.set_compression(CompressionMode.FILE, CompressionFormat.ZSTD)
    with bag:
        connection = bag.add_connection('/test', 'std_msgs/msg/Int8', typestore=store)
        bag.write(connection, 42, b'\x00')
        bag.write(connection, 666, b'\x01' * 4096)
    assert (path / 'metadata.yaml').exists()
    assert not (path / 'compress_file.db3').exists()
    assert (path / 'compress_file.db3.zstd').exists()

    path = tmp_path / 'compress_message'
    bag = Writer(path, version=Writer.VERSION_LATEST)
    bag.set_compression(CompressionMode.MESSAGE, CompressionFormat.ZSTD)
    with bag:
        connection = bag.add_connection('/test', 'std_msgs/msg/Int8', typestore=store)
        bag.write(connection, 42, b'\x00')
        bag.write(connection, 666, b'\x01' * 4096)
    assert (path / 'metadata.yaml').exists()
    assert (path / 'compress_message.db3').exists()
    assert size > (path / 'compress_message.db3').stat().st_size

    path = tmp_path / 'with_custom_data'
    bag = Writer(path, version=Writer.VERSION_LATEST)
    bag.open()
    bag.set_custom_data('key1', 'value1')
    with pytest.raises(WriterError, match='non-string value'):
        bag.set_custom_data('key1', 42)  # type: ignore[arg-type]
    bag.close()
    assert b'key1: value1' in (path / 'metadata.yaml').read_bytes()

    path = tmp_path / 'with_external_types'
    bag = Writer(path, version=Writer.VERSION_LATEST)
    with bag:
        connection = bag.add_connection(
            '/test',
            'std_msgs/msg/Int8',
            msgdef='msgdef',
            rihs01='RIHS00_hash',
        )
    assert b'RIHS00' in (path / 'metadata.yaml').read_bytes()
    assert b'RIHS01' not in (path / 'metadata.yaml').read_bytes()

    path = tmp_path / 'version9'
    bag = Writer(path, version=9)
    with bag:
        connection = bag.add_connection(
            '/test',
            'std_msgs/msg/Int8',
            msgdef='msgdef',
            rihs01='RIHS00_hash',
        )
    assert b'offered_qos_profiles: []' in (path / 'metadata.yaml').read_bytes()

    path = tmp_path / 'version8'
    bag = Writer(path, version=8)
    with bag:
        connection = bag.add_connection(
            '/test',
            'std_msgs/msg/Int8',
            msgdef='msgdef',
            rihs01='RIHS00_hash',
        )
    assert b"offered_qos_profiles: ''" in (path / 'metadata.yaml').read_bytes()


def test_failure_cases(tmp_path: Path) -> None:
    """Test writer failure cases."""
    store = get_typestore(Stores.LATEST)

    with pytest.raises(WriterError, match='exists'):
        _ = Writer(tmp_path, version=Writer.VERSION_LATEST)

    bag = Writer(tmp_path / 'race', version=Writer.VERSION_LATEST)
    (tmp_path / 'race').mkdir()
    with pytest.raises(WriterError, match='exists'):
        bag.open()

    bag = Writer(tmp_path / 'compress_after_open', version=Writer.VERSION_LATEST)
    bag.open()
    with pytest.raises(WriterError, match='already open'):
        bag.set_compression(CompressionMode.FILE, CompressionFormat.ZSTD)
    bag.close()

    bag = Writer(tmp_path / 'topic', version=Writer.VERSION_LATEST)
    with pytest.raises(WriterError, match='was not opened'):
        _ = bag.add_connection('/tf', 'tf2_msgs/msg/TFMessage')

    bag = Writer(tmp_path / 'write', version=Writer.VERSION_LATEST)
    with pytest.raises(WriterError, match='was not opened'):
        bag.write(
            Connection(
                1,
                '/tf',
                'tf2_msgs/msg/TFMessage',
                MessageDefinition(MessageDefinitionFormat.NONE, ''),
                '',
                0,
                ConnectionExtRosbag2('cdr', []),
                None,
            ),
            0,
            b'',
        )

    bag = Writer(tmp_path / 'topic', version=Writer.VERSION_LATEST)
    bag.open()
    _ = bag.add_connection('/tf', 'tf2_msgs/msg/TFMessage', typestore=store)
    _ = bag.add_connection(
        '/tf',
        'tf2_msgs/msg/TFMessage',
        typestore=store,
        serialization_format='ros1',
    )
    with pytest.raises(WriterError, match='only be added once'):
        _ = bag.add_connection('/tf', 'tf2_msgs/msg/TFMessage', typestore=store)
    bag.close()

    bag = Writer(tmp_path / 'notopic', version=Writer.VERSION_LATEST)
    bag.open()
    connection = Connection(
        1,
        '/tf',
        'tf_msgs/msg/tf2',
        MessageDefinition(MessageDefinitionFormat.NONE, ''),
        '',
        0,
        ConnectionExtRosbag2('cdr', []),
        None,
    )
    with pytest.raises(WriterError, match='unknown connection'):
        bag.write(connection, 42, b'\x00')
    bag.close()

    bag = Writer(tmp_path / 'bag', version=Writer.VERSION_LATEST)
    with bag, pytest.raises(WriterError, match='Cannot determine message definition'):
        _ = bag.add_connection('/foo', 'std_msgs/msg/Empty')
